"""
Jaime Sendra Berenguer-2020.
MLearner Machine Learning Library Extensions
Author:Jaime Sendra Berenguer<www.linkedin.com/in/jaisenbe>
License: MIT
"""


import numpy as np
import time

import tensorflow as tf
from tensorflow.keras import layers

from ..utils.keras import keras_checkpoint


"""
Based on:
https://www.udemy.com/course/procesamiento-del-lenguaje-natural/learn/lecture/21502260#overview

Author: Juan Gabriel Gomila
Course: "Procesamiento del Lenguaje Natural Moderno en Python.

Paper original: All you need is Attention https://arxiv.org/pdf/1706.03762.pdf
"""


class PositionalEncoding(layers.Layer):
    """
    Positional Encoding: Fórmula de la Codificación Posicional:
        $PE_{(pos,2i)} = \sin(pos/10000^{2i/dmodel})$
        $PE_{(pos,2i+1)} = \cos(pos/10000^{2i/dmodel})$
    """
    def __init__(self):
        super(PositionalEncoding, self).__init__()

    def get_angles(self, pos, i, d_model):  # pos: (seq_length, 1) i: (1, d_model)
        angles = 1 / np.power(10000., (2*(i//2)) / np.float32(d_model))
        return pos * angles  # (seq_length, d_model)

    def call(self, inputs):
        seq_length = inputs.shape.as_list()[-2]
        d_model = inputs.shape.as_list()[-1]
        angles = self.get_angles(np.arange(seq_length)[:, np.newaxis],
                                 np.arange(d_model)[np.newaxis, :],
                                 d_model)
        angles[:, 0::2] = np.sin(angles[:, 0::2])
        angles[:, 1::2] = np.cos(angles[:, 1::2])
        pos_encoding = angles[np.newaxis, ...]
        return inputs + tf.cast(pos_encoding, tf.float32)


"""
Attention
"""


def scaled_dot_product_attention(queries, keys, values, mask):
    """
    Cálculo de la Atención:

        $Attention(Q, K, V ) = \text{softmax}\left(\dfrac{QK^T}{\sqrt{d_k}}\right)V $
    """
    product = tf.matmul(queries, keys, transpose_b=True)

    keys_dim = tf.cast(tf.shape(keys)[-1], tf.float32)
    scaled_product = product / tf.math.sqrt(keys_dim)

    if mask is not None:
        scaled_product += (mask * -1e9)

    attention = tf.matmul(tf.nn.softmax(scaled_product, axis=-1), values)

    return attention


class MultiHeadAttention(layers.Layer):
    """
    Sub capa de atención de encabezado múltiple

    Parameters
    ----------
        nb_proj:
            Número de projecciones
    """
    def __init__(self, nb_proj):
        super(MultiHeadAttention, self).__init__()
        self.nb_proj = nb_proj

    def build(self, input_shape):
        self.d_model = input_shape[-1]
        assert self.d_model % self.nb_proj == 0

        self.d_proj = self.d_model // self.nb_proj

        self.query_lin = layers.Dense(units=self.d_model)
        self.key_lin = layers.Dense(units=self.d_model)
        self.value_lin = layers.Dense(units=self.d_model)

        self.final_lin = layers.Dense(units=self.d_model)

    def split_proj(self, inputs, batch_size):  # inputs: (batch_size, seq_length, d_model)
        shape = (batch_size,
                 -1,
                 self.nb_proj,
                 self.d_proj)
        splited_inputs = tf.reshape(inputs, shape=shape)  # (batch_size, seq_length, nb_proj, d_proj)
        return tf.transpose(splited_inputs, perm=[0, 2, 1, 3])  # (batch_size, nb_proj, seq_length, d_proj)

    def call(self, queries, keys, values, mask):
        batch_size = tf.shape(queries)[0]

        queries = self.query_lin(queries)
        keys = self.key_lin(keys)
        values = self.value_lin(values)

        queries = self.split_proj(queries, batch_size)
        keys = self.split_proj(keys, batch_size)
        values = self.split_proj(values, batch_size)

        attention = scaled_dot_product_attention(queries, keys, values, mask)

        attention = tf.transpose(attention, perm=[0, 2, 1, 3])

        concat_attention = tf.reshape(attention,
                                      shape=(batch_size, -1, self.d_model))

        outputs = self.final_lin(concat_attention)
        return outputs


class EncoderLayer(layers.Layer):
    """
    Capa de Codificación.

    Parameters
    ----------
        FFN_units:
            Número de capas densas.

        nb_proj:
            Número de projecciones.

        dropout_rate: Float
            Dropout parameter.
    """
    def __init__(self, FFN_units, nb_proj, dropout_rate):
        super(EncoderLayer, self).__init__()
        self.FFN_units = FFN_units
        self.nb_proj = nb_proj
        self.dropout_rate = dropout_rate

    def build(self, input_shape):
        self.d_model = input_shape[-1]

        self.multi_head_attention = MultiHeadAttention(self.nb_proj)
        self.dropout_1 = layers.Dropout(rate=self.dropout_rate)
        self.norm_1 = layers.LayerNormalization(epsilon=1e-6)

        self.dense_1 = layers.Dense(units=self.FFN_units, activation="relu")
        self.dense_2 = layers.Dense(units=self.d_model)
        self.dropout_2 = layers.Dropout(rate=self.dropout_rate)
        self.norm_2 = layers.LayerNormalization(epsilon=1e-6)

    def call(self, inputs, mask, training):
        attention = self.multi_head_attention(inputs,
                                              inputs,
                                              inputs,
                                              mask)
        attention = self.dropout_1(attention, training=training)
        attention = self.norm_1(attention + inputs)

        outputs = self.dense_1(attention)
        outputs = self.dense_2(outputs)
        outputs = self.dropout_2(outputs, training=training)
        outputs = self.norm_2(outputs + attention)

        return outputs


class Encoder(layers.Layer):
    """
    Codificación.

    Parameters
    ----------
        nb_layers:
            Número de capas del transformer.

        FFN_units:
            Número de capas densas.

        nb_proj:
            Número de projecciones.

        dropout_rate: Float
            Dropout parameter.

        vocab_size:
            Vocabulary size of the algorithm input text.

        d_model: int
            Profundidad del embeding.
    """
    def __init__(self,
                 nb_layers,
                 FFN_units,
                 nb_proj,
                 dropout_rate,
                 vocab_size,
                 d_model,
                 name="encoder"):
        super(Encoder, self).__init__(name=name)
        self.nb_layers = nb_layers
        self.d_model = d_model

        self.embedding = layers.Embedding(vocab_size, d_model)
        self.pos_encoding = PositionalEncoding()
        self.dropout = layers.Dropout(rate=dropout_rate)
        self.enc_layers = [EncoderLayer(FFN_units,
                                        nb_proj,
                                        dropout_rate)
                           for _ in range(nb_layers)]

    def call(self, inputs, mask, training):
        outputs = self.embedding(inputs)
        outputs *= tf.math.sqrt(tf.cast(self.d_model, tf.float32))
        outputs = self.pos_encoding(outputs)
        outputs = self.dropout(outputs, training)

        for i in range(self.nb_layers):
            outputs = self.enc_layers[i](outputs, mask, training)

        return outputs


class DecoderLayer(layers.Layer):
    """
    Capa de Descodificación.

    Parameters
    ----------
        FFN_units:
            Número de capas densas.

        nb_proj:
            Número de projecciones.

        dropout_rate: Float
            Dropout parameter.
    """
    def __init__(self, FFN_units, nb_proj, dropout_rate):
        super(DecoderLayer, self).__init__()
        self.FFN_units = FFN_units
        self.nb_proj = nb_proj
        self.dropout_rate = dropout_rate

    def build(self, input_shape):
        self.d_model = input_shape[-1]

        # Self multi head attention
        self.multi_head_attention_1 = MultiHeadAttention(self.nb_proj)
        self.dropout_1 = layers.Dropout(rate=self.dropout_rate)
        self.norm_1 = layers.LayerNormalization(epsilon=1e-6)

        # Multi head attention combinado con la salida del encoder
        self.multi_head_attention_2 = MultiHeadAttention(self.nb_proj)
        self.dropout_2 = layers.Dropout(rate=self.dropout_rate)
        self.norm_2 = layers.LayerNormalization(epsilon=1e-6)

        # Feed foward
        self.dense_1 = layers.Dense(units=self.FFN_units,
                                    activation="relu")
        self.dense_2 = layers.Dense(units=self.d_model)
        self.dropout_3 = layers.Dropout(rate=self.dropout_rate)
        self.norm_3 = layers.LayerNormalization(epsilon=1e-6)

    def call(self, inputs, enc_outputs, mask_1, mask_2, training):
        attention = self.multi_head_attention_1(inputs,
                                                inputs,
                                                inputs,
                                                mask_1)
        attention = self.dropout_1(attention, training)
        attention = self.norm_1(attention + inputs)

        attention_2 = self.multi_head_attention_2(attention,
                                                  enc_outputs,
                                                  enc_outputs,
                                                  mask_2)
        attention_2 = self.dropout_2(attention_2, training)
        attention_2 = self.norm_2(attention_2 + attention)

        outputs = self.dense_1(attention_2)
        outputs = self.dense_2(outputs)
        outputs = self.dropout_3(outputs, training)
        outputs = self.norm_3(outputs + attention_2)

        return outputs


class Decoder(layers.Layer):
    """
    Descodificación.

    Parameters
    ----------
        nb_layers:
            Número de capas del transformer.

        FFN_units:
            Número de capas densas.

        nb_proj:
            Número de projecciones.

        dropout_rate: Float
            Dropout parameter.

        vocab_size:
            Vocabulary size of the algorithm input text.

        d_model: int
            Profundidad del embeding.
    """
    def __init__(self,
                 nb_layers,
                 FFN_units,
                 nb_proj,
                 dropout_rate,
                 vocab_size,
                 d_model,
                 name="decoder"):
        super(Decoder, self).__init__(name=name)
        self.d_model = d_model
        self.nb_layers = nb_layers

        self.embedding = layers.Embedding(vocab_size, d_model)
        self.pos_encoding = PositionalEncoding()
        self.dropout = layers.Dropout(rate=dropout_rate)

        self.dec_layers = [DecoderLayer(FFN_units,
                                        nb_proj,
                                        dropout_rate)
                           for _ in range(nb_layers)]

    def call(self, inputs, enc_outputs, mask_1, mask_2, training):
        outputs = self.embedding(inputs)
        outputs *= tf.math.sqrt(tf.cast(self.d_model, tf.float32))
        outputs = self.pos_encoding(outputs)
        outputs = self.dropout(outputs, training)

        for i in range(self.nb_layers):
            outputs = self.dec_layers[i](outputs,
                                         enc_outputs,
                                         mask_1,
                                         mask_2,
                                         training)
        return outputs


class Transformer(tf.keras.Model):
    """
    Transformer

    Parameters
    ----------
        vocab_size_enc:
            Tamaño vocabulario entrada del transformer.

        vocab_size_dec:
            Tamaño vocabulario salida del transformer.

        nb_layers:
            Número de capas del transformer.

        FFN_units:
            Número de capas densas.

        nb_proj:
            Número de projecciones.

        dropout_rate: Float
            Dropout parameter.

        vocab_size:
            Vocabulary size of the algorithm input text.

        d_model: int
            Profundidad del embeding.

        name: str
            Nombre asociado a la instancia de la clase Transformer.

    Attributes
    -----------
        encoder : tf.keras.layers.Embedding
            encoder custom Layer

        decoder : tf.keras.layers.Embedding
            decoder custom Layer

    Examples:
    ---------
        ```python

        BATCH_SIZE = 64
        BUFFER_SIZE = 20000

        dataset = tf.data.Dataset.from_tensor_slices((inputs, outputs))

        dataset = dataset.cache()
        dataset = dataset.shuffle(BUFFER_SIZE).batch(BATCH_SIZE)
        dataset = dataset.prefetch(tf.data.experimental.AUTOTUNE)

        tf.keras.backend.clear_session()

        # Hiper Parámetros
        D_MODEL = 128 # 512
        NB_LAYERS = 4 # 6
        FFN_UNITS = 512 # 2048
        NB_PROJ = 8 # 8
        DROPOUT_RATE = 0.1 # 0.1

        model_Transformer = Transformer(vocab_size_enc=VOCAB_SIZE_EN,
                                vocab_size_dec=VOCAB_SIZE_ES,
                                d_model=D_MODEL,
                                nb_layers=NB_LAYERS,
                                FFN_units=FFN_UNITS,
                                nb_proj=NB_PROJ,
                                dropout_rate=DROPOUT_RATE)

        Transformer_train(model_Transformer,
                        dataset,
                        d_model=D_MODEL,
                        train=TRAIN,
                        epochs=1,
                        checkpoint_path="ckpt/",
                        max_to_keep=5)

        # Evaluate
        def evaluate(inp_sentence):
            inp_sentence = \
                [VOCAB_SIZE_EN-2] + processor_en.tokenizer.encode(inp_sentence) + [VOCAB_SIZE_EN-1]
            enc_input = tf.expand_dims(inp_sentence, axis=0)

            output = tf.expand_dims([VOCAB_SIZE_ES-2], axis=0)

            for _ in range(MAX_LENGTH):
                predictions = model_Transformer(enc_input, output, False) #(1, seq_length, VOCAB_SIZE_ES)

                prediction = predictions[:, -1:, :]

                predicted_id = tf.cast(tf.argmax(prediction, axis=-1), tf.int32)

                if predicted_id == VOCAB_SIZE_ES-1:
                    return tf.squeeze(output, axis=0)

                output = tf.concat([output, predicted_id], axis=-1)

            return tf.squeeze(output, axis=0)

        def translate(sentence):
            output = evaluate(sentence).numpy()

            predicted_sentence = processor_es.tokenizer.decode(
                [i for i in output if i < VOCAB_SIZE_ES-2]
            )

            print("Entrada: {}".format(sentence))
            print("Traducción predicha: {}".format(predicted_sentence))

        translate("This is a problem we have to solve.")

            >>> Entrada: This is a problem we have to solve.
            >>> Traducción predicha: Es un problema que debemos resolver.

        ```
    """
    def __init__(self,
                 vocab_size_enc,
                 vocab_size_dec,
                 d_model,
                 nb_layers,
                 FFN_units,
                 nb_proj,
                 dropout_rate,
                 name="transformer"):
        super(Transformer, self).__init__(name=name)

        self.encoder = Encoder(nb_layers,
                               FFN_units,
                               nb_proj,
                               dropout_rate,
                               vocab_size_enc,
                               d_model)
        self.decoder = Decoder(nb_layers,
                               FFN_units,
                               nb_proj,
                               dropout_rate,
                               vocab_size_dec,
                               d_model)
        self.last_linear = layers.Dense(units=vocab_size_dec, name="lin_ouput")

    def create_padding_mask(self, seq):  # seq: (batch_size, seq_length)
        mask = tf.cast(tf.math.equal(seq, 0), tf.float32)
        return mask[:, tf.newaxis, tf.newaxis, :]

    def create_look_ahead_mask(self, seq):
        seq_len = tf.shape(seq)[1]
        look_ahead_mask = 1 - tf.linalg.band_part(tf.ones((seq_len, seq_len)), -1, 0)
        return look_ahead_mask

    def call(self, enc_inputs, dec_inputs, training):
        enc_mask = self.create_padding_mask(enc_inputs)
        dec_mask_1 = tf.maximum(
            self.create_padding_mask(dec_inputs),
            self.create_look_ahead_mask(dec_inputs)
        )
        dec_mask_2 = self.create_padding_mask(enc_inputs)

        enc_outputs = self.encoder(enc_inputs, enc_mask, training)
        dec_outputs = self.decoder(dec_inputs,
                                   enc_outputs,
                                   dec_mask_1,
                                   dec_mask_2,
                                   training)

        outputs = self.last_linear(dec_outputs)

        return outputs


class CustomSchedule_transformer(tf.keras.optimizers.schedules.LearningRateSchedule):
    """
    Custom Learning Rate Schedule:
        (5.3.) https://arxiv.org/pdf/1706.03762.pdf

    Parameters
    ----------
        d_model: int
            Número de projecciones

        warmup_steps: int

    """
    def __init__(self, d_model, warmup_steps=4000):
        super(CustomSchedule_transformer, self).__init__()

        self.d_model = tf.cast(d_model, tf.float32)
        self.warmup_steps = warmup_steps

    def __call__(self, step):
        arg1 = tf.math.rsqrt(step)
        arg2 = step * (self.warmup_steps**-1.5)

        return tf.math.rsqrt(self.d_model) * tf.math.minimum(arg1, arg2)


def loss_function(target, pred):
    """
    Función de cáculo de pérdidas.

    Se calcula a partir de elementos no nulos, por ello
    extraemos elementos del padding.
    """
    # Mascara para deshacerme de los "ceros" del padding
    mask = tf.math.logical_not(tf.math.equal(target, 0))
    loss_object = tf.keras.losses.SparseCategoricalCrossentropy(from_logits=True,
                                                                reduction="none")
    loss_ = loss_object(target, pred)

    # Equiparar el tipo de dato de la mascara al de la perdida
    mask = tf.cast(mask, dtype=loss_.dtype)

    # Multiplicamos elemento a elemento.
    loss_ *= mask

    # Media por bloques (media de los resultados no nulos)
    return tf.reduce_mean(loss_)


def Transformer_train(Transformer,
                        dataset,
                        d_model,
                        epochs,
                        train=True,
                        beta_1=0.9,
                        beta_2=0.98,
                        epsilon=1e-9,
                        checkpoint_path="ckpt/",
                        max_to_keep=5):
    """
    Entrenamiento Transformer Customizado
    """
    # Custom Learning Rate Schedule
    leaning_rate = CustomSchedule_transformer(d_model)
    # Loss function
    train_loss = tf.keras.metrics.Mean(name="train_loss")
    train_accuracy = tf.keras.metrics.SparseCategoricalAccuracy(name="train_accuracy")
    # Optimizador
    optimizer = tf.keras.optimizers.Adam(leaning_rate,
                                            beta_1=beta_1,
                                            beta_2=beta_2,
                                            epsilon=epsilon)
    # Checkpoints
    ckpt_manager = keras_checkpoint(Transformer,
                                    optimizer,
                                    checkpoint_path="ckpt/",
                                    max_to_keep=5)
    # Grafo estatico
    @tf.function
    def train_step(enc_inputs, dec_inputs, dec_outputs_real):
        with tf.GradientTape() as tape:
            predictions = Transformer(enc_inputs, dec_inputs, True)
            loss = loss_function(dec_outputs_real, predictions)

        gradients = tape.gradient(loss, Transformer.trainable_variables)
        optimizer.apply_gradients(zip(gradients, Transformer.trainable_variables))

        train_accuracy(dec_outputs_real, predictions)
        return loss

    # Bucle de Entrenamiento
    if train:
        for epoch in range(epochs):
            print("Inicio del epoch {}".format(epoch+1))
            start = time.time()

            train_loss.reset_states()
            train_accuracy.reset_states()

            for (batch, (enc_inputs, targets)) in enumerate(dataset):
                dec_inputs = targets[:, :-1]
                dec_outputs_real = targets[:, 1:]

                loss = train_step(enc_inputs, dec_inputs, dec_outputs_real)
                train_loss(loss)

                if batch % 50 == 0:
                    print("Epoch {} Lote {} Pérdida {:.4f} Precisión {:.4f}".format(
                        epoch+1, batch, train_loss.result(), train_accuracy.result()))

            ckpt_save_path = ckpt_manager.save()
            print("Guardando checkpoint para el epoch {} en {}".format(epoch+1,
                                                                        ckpt_save_path))
            print("Tiempo que ha tardado 1 epoch: {} segs\n".format(time.time() - start))



"""
Jaime Sendra Berenguer-2020.
MLearner Machine Learning Library Extensions
Author:Jaime Sendra Berenguer<www.linkedin.com/in/jaisenbe>
License: MIT
"""

import os
import pandas as pd

this_dir, this_filename = os.path.split(__file__)
DATA_PATH = os.path.join(this_dir, "data", "wine.csv")


def wine_data():
    """
    Wine dataset.

    Source: https://archive.ics.uci.edu/ml/datasets/Wine
    Number of samples: 178
    Class labels: {0, 1, 2}, distribution: [59, 71, 48]

    Data Set Information:

    These data are the results of a chemical analysis of wines grown
    in the same region in Italy but derived from three different cultivars.
    The analysis determined the quantities of 13 constituents found in each
    of the three types of wines.

    The attributes are (dontated by Riccardo Leardi, riclea@anchem.unige.it)

        - 1) Alcohol
        - 2) Malic acid
        - 3) Ash
        - 4) Alcalinity of ash
        - 5) Magnesium
        - 6) Total phenols
        - 7) Flavanoids
        - 8) Nonflavanoid phenols
        - 9) Proanthocyanins
        - 10) Color intensity
        - 11) Hue
        - 12) OD280/OD315 of diluted wines
        - 13) Proline

    In a classification context, this is a well posed problem with "well behaved"
    class structures. A good data set for first testing of a new classifier,
    but not very challenging.

    Returns
    --------
    X, y : [n_samples, n_features], [n_class_labels]
        X is the feature matrix with 178 wine samples as rows
        and 13 feature columns.
        y is a 1-dimensional array of the 3 class labels 0, 1, 2

    Examples
    --------
    For usage examples, please see
    https://jaisenbe58r.github.io/MLearner/user_guide/data/wine_data

    adapted from
    https://github.com/rasbt/mlxtend/blob/master/mlxtend/data/wine.py
    Author: Sebastian Raschka <sebastianraschka.com>
    License: BSD 3 clause

    """
    header = ['alcohol', 'malic acid', 'ash', 'ash alcalinity',
                'magnesium', 'total phenols', 'flavanoids',
                'nonflavanoid phenols', 'proanthocyanins',
                'color intensity', 'hue', 'OD280/OD315 of diluted wines',
                'proline']
    data_csv = pd.read_csv(DATA_PATH, delimiter=',', names=header)
    X = data_csv[data_csv.columns[:-1]]
    y = data_csv[data_csv.columns[-1]].astype(int)
    return X, y




"""Jaime Sendra Berenguer-2020.
MLearner Machine Learning Library Extensions
Author:Jaime Sendra Berenguer<www.linkedin.com/in/jaisenbe>
License: MIT
"""

import numpy as np
import pandas as pd
from scipy import stats


def data_uniform(a, b, n):
    """Generate a Uniform data distribution.

    Attributes
    ----------
    a : `int` or `float`
        manimum value of the entire dataset.
    b : `int` or `float`
        maximum value of the entire dataset.
    n : `int`
        number of data in the dataset.

    Returns
    -------
    data : Uniform data distribution.

    Examples
    --------
    For usage examples, please see
    https://jaisenbe58r.github.io/MLearner/user_guide/data/data_uniform/


    """
    if not isinstance(a, int) and not isinstance(a, float):
        raise TypeError("Invalid type for a : {}".format(type(a)))
    if not isinstance(b, int) and not isinstance(b, float):
        raise TypeError("Invalid type for a : {}".format(type(b)))
    if not isinstance(n, int):
        raise TypeError("Invalid type for a : {}".format(type(n)))

    if n <= 0:
        raise NameError("'n' must be an integer greater than zero")

    data = np.random.uniform(a, b, n)
    return data


def data_normal(mu=0, sd=1, n=100):
    """Generate a Normal data distribution.

    Attributes
    ----------
    mu : `int` or `float`
        mean value.
    sd : `int` or `float`
        standard deviation.
    n : `int`
        number of data in the dataset.

    Returns
    -------
    data : Uniform data distribution.

    Examples
    --------
    For usage examples, please see
    https://jaisenbe58r.github.io/MLearner/user_guide/data/data_normal/


    """
    if not isinstance(mu, int) and not isinstance(mu, float):
        raise TypeError("Invalid type for a : {}".format(type(mu)))
    if not isinstance(sd, int) and not isinstance(sd, float):
        raise TypeError("Invalid type for a : {}".format(type(sd)))
    if not isinstance(n, int):
        raise TypeError("Invalid type for a : {}".format(type(n)))

    if n <= 0:
        raise NameError("'n' must be an integer greater than zero")

    data = mu + sd * np.random.randn(n)
    return data


def data_gamma(a=5, n=100):
    """Generate a Gamma data distribution.

    Attributes
    ----------
    a : `int` or `float`
        Parameter form.
    n : `int`
        number of data in the dataset.

    Returns
    -------
    data : Uniform data distribution.

    Examples
    --------
    For usage examples, please see
    https://jaisenbe58r.github.io/MLearner/user_guide/data/data_gamma/


    """
    if not isinstance(a, int) and not isinstance(a, float):
        raise TypeError("Invalid type for a : {}".format(type(a)))
    if not isinstance(n, int):
        raise TypeError("Invalid type for a : {}".format(type(n)))

    if n <= 0:
        raise NameError("'n' must be an integer greater than zero")

    gamma = stats.gamma(a)
    data = gamma.rvs(n)
    return data


def create_dataset(config, n):
    """Generate a Dataset.

    Attributes
    ----------
    config : `dict`
        Dictionary for dataset configuration:
            p.e.:
            ```
            dict = {
                'A' : data_uniform(0, 1, n),
                'B' : data_normal(n),
                'C' : data_normal(mu=5, sd=2, n=n),
                'D' : data_gamma(a=5, n=n)
            }
            ```
    n : `int`
        number of data in the dataset.

    Returns
    -------
    data : Dataset

    Examples
    --------
    For usage examples, please see
    https://jaisenbe58r.github.io/MLearner/user_guide/data/create_dataset/

    """
    if not isinstance(config, dict):
        raise TypeError("Invalid type for 'config' : {}".format(type(config)))

    if not isinstance(n, int):
        raise TypeError("Invalid type for a : {}".format(type(n)))

    if n <= 0:
        raise NameError("'n' must be an integer greater than zero")

    data = pd.DataFrame(config)
    return data




"""Jaime Sendra Berenguer-2020.
MLearner Machine Learning Library Extensions
Author:Jaime Sendra Berenguer <www.linkedin.com/in/jaisenbe>
License: MIT
"""


from sklearn.preprocessing import LabelEncoder, StandardScaler
from .scaling import minmax_scaling
from .mean_centering import MeanCenterer
from .droper import FeatureDropper, DropOutliers
from .replace_na import FillNaTransformer_forward, FillNaTransformer_backward, FillNaTransformer_value, FillNaTransformer_all, FillNaTransformer_any, FillNaTransformer_median, FillNaTransformer_mean, FillNaTransformer_idmax
from .log_skewed import FixSkewness
from .one_hot_encoder import OneHotEncoder, CategoricalEncoder
from .replace_categorical import ReplaceTransformer, ReplaceMulticlass
from .extract_target import ExtractCategories
from .base_preprocess import DataExploratory, DataAnalyst
from .feature_selector import FeatureSelector, DataFrameSelector, DropFeatures, CopyFeatures
from .reduce_feature import PCA_selector, LDA_selector, PCA_add, LDA_add
from .perfomance_transformers import OrientationClassTransformer, MFD_OrientationClassTransformer
from .perfomance_transformers import ClassTransformer_value, Keep

__all__ = ["minmax_scaling", "MeanCenterer", "FeatureDropper",
            "FillNaTransformer_median", "FillNaTransformer_mean",
            "FillNaTransformer_idmax", "FillNaTransformer_any",
            "FillNaTransformer_all", "FillNaTransformer_value",
            "FillNaTransformer_backward", "FillNaTransformer_forward",
            "FixSkewness", "OneHotEncoder", "DropOutliers", "ReplaceTransformer",
            "ReplaceMulticlass", "ExtractCategories", "DataExploratory",
            "DataAnalyst", "LabelEncoder", "StandardScaler",
            "FeatureSelector", "DataFrameSelector", "CategoricalEncoder",
            "PCA_selector", "LDA_selector", "PCA_add", "LDA_add",
            "DropFeatures", "CopyFeatures",
            "OrientationClassTransformer", "MFD_OrientationClassTransformer"
            "ClassTransformer_value", "Keep"]




"""Jaime Sendra Berenguer-2020.
MLearner Machine Learning Library Extensions
Author:Jaime Sendra Berenguer<www.linkedin.com/in/jaisenbe>
License: MIT
"""

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

import seaborn as sns
from sklearn.model_selection import GridSearchCV

from sklearn.model_selection import StratifiedKFold, RandomizedSearchCV
from sklearn.metrics import make_scorer, accuracy_score, mean_squared_error
from sklearn.metrics import classification_report

from mlearner.evaluation import EvaluationModels

from tqdm import tqdm

import warnings
warnings.filterwarnings("ignore")


class Training(EvaluationModels):
    def __init__(self, model, random_state=99):
        self.model = model
        self.random_state = random_state
        super().__init__(self.model, self.random_state)

    @classmethod
    def add_model(cls, model, random_state=99):
        """
        Incorporar  modelo en la clase
        """
        return cls(model, random_state)

    def get_model(self):
        return self.model

    def KFold_CrossValidation(self, model, X, y, n_splits=10, ROC=False, shuffle=True, mute=False,
                                logdir_report="", display=True, save_image=True, verbose=0):
        """
        Validacion cruzada respecto a "n_plits" del KFolds.
        """
        y, targets = self.replace_multiclass(y)
        kf = StratifiedKFold(n_splits=n_splits, shuffle=shuffle, random_state=self.random_state)

        results = pd.DataFrame(columns=['training_score', 'test_score'])
        fprs, tprs, scores = [], [], []

        if not ROC:
            for (train, test), i in tqdm(zip(kf.split(X, y), range(n_splits))):
                model.fit(X.iloc[train], y.iloc[train])

                predictions_train = model.predict(X.iloc[train])
                predictions_test = model.predict(X.iloc[test])

                score_train = accuracy_score(y.iloc[train], predictions_train)
                score_test = accuracy_score(y.iloc[test], predictions_test)

                scores.append((score_train, score_test))
                if not mute:
                    print("  * Fold {0} accuracy: {1}".format(i+1, score_test))

            resultados = pd.DataFrame(scores, columns=['Score Train', 'Score Test'])
            score_general_test = resultados['Score Test'].mean()

        else:
            for (train, test), i in tqdm(zip(kf.split(X, y), range(n_splits))):
                model.fit(X.iloc[train], y.iloc[train])
                _, _, auc_score_train = self.compute_roc_auc(model, train, X, y)
                fpr, tpr, auc_score = self.compute_roc_auc(model, test, X, y)

                predictions_train = model.predict(X.iloc[train])
                predictions_test = model.predict(X.iloc[test])

                score_train = accuracy_score(y.iloc[train], predictions_train)
                score_test = accuracy_score(y.iloc[test], predictions_test)

                scores.append((auc_score_train, auc_score, score_train, score_test))
                fprs.append(fpr)
                tprs.append(tpr)

            self.create_ROC_pro(fprs, tprs, X, y, targets=targets, logdir_report=logdir_report,
                                    display=display, save_image=save_image)
            resultados = pd.DataFrame(scores, columns=['AUC Train', 'AUC Test',
                                                        'Acc Train', 'Acc Test'])
            score_general_test = resultados['Acc Test'].mean()

        if not mute:
            print("\n")
            print("----> Resultado de la validacion cruzada: {:.3f}%".format(score_general_test*100))
            if ROC:
                print("----> Resultado de la validacion cruzada AUC: {:.3f}%".format(
                                                                resultados['AUC Test'].mean()*100))
        return resultados, score_general_test

    def FineTune(self, model, X, y, params, refit='Accuracy', cv=3, verbose=0, randomized=True, n_iter=100, mute=False):
        """
        Tecnica de Ajuste fino de hiperparametros.

        Model: Modelo a Optimizar.

        params: diccionario de parametros con el grid.

        scoring: Metricas. scoring = {'AUC': 'roc_auc', 'Accuracy': acc_scorer}
            - Anotador de metricas: acc_score = make_scorer(accuracy_score, mean_squared_error)

        refit: Metrica de importancia para optimizar el modelo'Accuracy'
        """
        acc_scorer = make_scorer(accuracy_score, mean_squared_error)
        scoring = {
            "AUC": "roc_auc",
            "Accuracy": acc_scorer
            }

        self._pre_model = model

        self.scoring = scoring

        if randomized:
            self.grid_obj = RandomizedSearchCV(model, params, n_iter=n_iter, cv=cv, verbose=verbose,
                                                random_state=self.random_state, n_jobs=-1)
        else:
            self.grid_obj = GridSearchCV(model, params, scoring=scoring, cv=cv, refit=refit,
                                            return_train_score=True, verbose=verbose)

        self.grid_obj.fit(X, y)
        _best_Parameters = self.grid_obj.best_params_

        _results_df = pd.DataFrame(self.grid_obj.cv_results_)

        if not mute:
            print('=='*38)
            print("Mejores parametros obtenidos en el Fine-Tune:")
            print(_best_Parameters)
            print('=='*38)

        return _best_Parameters, _results_df

    def eval_train(self, model, X, y, name="Performance"):

        predictions = model.predict(X)
        errors = abs(predictions - y)
        mape = 100 * np.mean(errors / y)
        accuracy = 100 - mape
        print('=='*50)
        print('Model ' + name)
        print('Average Error: {:0.4f} degrees.'.format(np.mean(errors)))
        print('Accuracy = {:0.2f}%.'.format(accuracy))
        print('=='*50)
        print("\n")

        return accuracy

    def replace_multiclass(self, targets):

        _unic = targets.unique().tolist()
        _unic = list(np.sort(_unic))
        _remp = np.arange(0, len(_unic)).tolist()
        return targets.replace(_unic, _remp), _unic

    def eval_FineTune(self, X, y):
        """
        https://towardsdatascience.com/hyperparameter-tuning-the-random-forest-in-python-using-scikit-learn-28d2aa77dd74
        https://github.com/WillKoehrsen/Machine-Learning-Projects/blob/master/random_forest_explained/Improving%20Random%20Forest%20Part%202.ipynb

        """
        base_accuracy = self.eval_train(self._pre_model, X, y, name='FineTune')
        new_accuracy = self.eval_train(self.model, X, y, name='Inicial')
        print('Mejora del {:0.2f}%.'.format(100 * (new_accuracy - base_accuracy)/base_accuracy))

    def GridSearchCV_Evaluating(self, model, param, max_param, min_score=0.5):
        """
        https://scikit-learn.org/stable/auto_examples/model_selection/plot_multi_metric_evaluation.html#sphx-glr-auto-examples-model-selection-plot-multi-metric-evaluation-py
        """
        plt.figure(figsize=(13, 13))
        plt.title("GridSearchCV evaluating using multiple scorers simultaneously", fontsize=16)

        plt.xlabel("n_estimators")
        plt.ylabel("Score")

        ax = plt.gca()
        ax.set_xlim(0, max_param)
        ax.set_ylim(min_score, 1)

        results = self.grid_obj.cv_results_

        # Get the regular numpy array from the MaskedArray
        X_axis = np.array(results[param[0]].data, dtype=float)

        for scorer, color in tqdm(zip(sorted(self.scoring), ['g', 'k'])):
            for sample, style in (('train', '--'), ('test', '-')):
                sample_score_mean = results['mean_%s_%s' % (sample, scorer)]
                sample_score_std = results['std_%s_%s' % (sample, scorer)]
                ax.fill_between(X_axis, sample_score_mean - sample_score_std,
                                sample_score_mean + sample_score_std,
                                alpha=0.1 if sample == 'test' else 0, color=color)
                ax.plot(X_axis, sample_score_mean, style, color=color,
                        alpha=1 if sample == 'test' else 0.7,
                        label="%s (%s)" % (scorer, sample))

            best_index = np.nonzero(results['rank_test_%s' % scorer] == 1)[0][0]
            best_score = results['mean_test_%s' % scorer][best_index]

            # Plot a dotted vertical line at the best score for that scorer marked by x
            ax.plot([X_axis[best_index], ] * 2, [0, best_score],
                    linestyle='-.', color=color, marker='x', markeredgewidth=3, ms=8)

            # Annotate the best score for that scorer
            ax.annotate("%0.2f" % best_score,
                        (X_axis[best_index], best_score + 0.005))

        plt.legend(loc="best")
        plt.grid(False)
        plt.show()

    def evaluate(self, model, X, y):
        """
        Evalucion del modelo Fine-Tune
        """
        # print(model.get_params())
        print('=='*35)
        predictions = model.predict(X)
        preds = [np.argmax(line) for line in predictions]
        report = classification_report(y, preds)
        score = accuracy_score(y_true=y, y_pred=preds)

        print(report)
        print('=='*35)
        print("{} {:0.2f}%".format("Accuracy Score evaluation : ", score*100))

    def heatmap_params(self, parameters, results_df, metric='mean_test_Accuracy'):
        """
        parametres a relacionar:
            parameters = ["n_estimators", "min_samples_split"]
        """
        x = len(self.params[parameters[0]])
        y = len(self.params[parameters[1]])

        scores = np.array(results_df[metric].values).reshape(x, y)

        sns.heatmap(scores, annot=True,
                        xticklabels=self.params[parameters[0]],
                        yticklabels=self.params[parameters[1]])

    def features_important(self, X, y, logdir="", display=True, save_image=False):
        """
        Explorar las features mas significativas
        """
        # features:
        features = X.columns.tolist()
        # Get numerical feature importances
        importances = list(self.model.feature_importances_)
        # List of tuples with variable and importance
        feature_importances = [(feature, round(importance, 2)) for feature, importance in zip(features, importances)]
        # Sort the feature importances by most important first
        feature_importances = sorted(feature_importances, key=lambda x: x[1], reverse=True)
        # Print out the feature and importances
        [print('Variable: {:20} Importance: {}'.format(*pair)) for pair in feature_importances]

        if display or save_image:
            # Set the style
            plt.style.use('seaborn')
            # list of x locations for plotting
            x_values = list(range(len(importances)))
            # Make a bar chart
            plt.bar(x_values, importances, orientation='vertical')
            # Tick labels for x axis
            plt.xticks(x_values, features, rotation='vertical')
            # Axis labels and title
            plt.ylabel('Importancia')
            plt.xlabel('Variables')
            plt.title('Importancia de las variables')

        if display:
            plt.show()
        if save_image:
            plt.savefig(logdir)




"""Jaime Sendra Berenguer-2020.
MLearner Machine Learning Library Extensions
Author:Jaime Sendra Berenguer<www.linkedin.com/in/jaisenbe>
License: MIT
"""

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin


class MeanCenterer(BaseEstimator, TransformerMixin):

    """Column centering of pandas Dataframe.

    Attributes
    ----------
    col_means:  numpy.ndarray [n_columns] or pandas [n_columns]
    mean values for centering after fitting the MeanCenterer object.

    Examples
    --------
    For usage examples, please see
    https://jaisenbe58r.github.io/MLearner/user_guide/preprocessing/MeanCenterer/

    adapted from
    https://github.com/rasbt/mlxtend/blob/master/mlxtend/preprocessing/mean_centering.py
    Author: Sebastian Raschka <sebastianraschka.com>
    License: BSD 3 clause

    """
    def __init__(self, columns=None):
        """Init Mean Center."""
        if columns is not None:
            if isinstance(columns, list) or isinstance(columns, tuple):
                self.columns = columns
            else:
                raise NameError("Invalid type {}".format(type(columns)))
        else:
            self.columns = columns

    def fit(self, X, y=None):
        """Gets the column means for mean centering.

        Parameters
        ----------
        X : {Dataframe}, shape = [n_samples, n_features]
            Dataframe, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        --------
        self
        """
        if isinstance(X, pd.core.frame.DataFrame):
            try:
                _test = X.astype(np.float32)
                del(_test)
            except ValueError:
                raise NameError("Null or categorical variables are not allowed: {}".format(X.dtypes))

            self.col_means = X.mean()

        elif isinstance(X, np.ndarray):
            try:
                _test = np.mean(X)
                del(_test)
            except TypeError:
                raise NameError("Null or categorical variables are not allowed")

            _X_fl = X.astype('float')
            self.col_means = _X_fl.mean(axis=0)

        else:
            raise NameError("Invalid type {}".format(type(X)))

        return self

    def transform(self, X):
        """Centers a pandas.

        Parameters
        ----------
        X : {Dataframe}, shape = [n_samples, n_features]
            Dataframe of samples, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        -------
        X_transform : {DAtaframe}, shape = [n_samples, n_features]
            A copy of the input Dataframe with the columns centered.

        """
        if not hasattr(self, "col_means"):
            raise AttributeError("MeanCenterer has not been fitted, yet.")

        if isinstance(X, pd.core.frame.DataFrame):
            X_transform = X.copy()
            # try:
            for i in range(X.shape[1]):
                X_transform[i] = np.apply_along_axis(func1d=lambda x: (x - X.mean()),
                                                        axis=1, arr=X)
            # except:
            #     raise NameError("Invalid shape {}".format(X.shape))
        elif isinstance(X, np.ndarray):
            X_transform = np.apply_along_axis(func1d=lambda x: x - self.col_means,
                                                axis=1, arr=X)
        else:
            raise NameError("Invalid type {}".format(type(X)))
        return X_transform




"""Jaime Sendra Berenguer-2020.
MLearner Machine Learning Library Extensions
Author:Jaime Sendra Berenguer<www.linkedin.com/in/jaisenbe>
License: MIT
"""

import numpy as np
import pandas as pd
from scipy.stats import skew
from sklearn.base import BaseEstimator, TransformerMixin


class FixSkewness(BaseEstimator, TransformerMixin):

    """This transformer applies log to skewed features.

    Attributes
    ----------
    columns:  npandas [n_columns]

    Examples
    --------
    For usage examples, please see:
    https://jaisenbe58r.github.io/MLearner/user_guide/preprocessing/FixSkewness/


    """
    def __init__(self, columns=None, drop=True):
        """Init log skewed."""
        if columns is not None:
            if isinstance(columns, list) or isinstance(columns, tuple):
                self.columns = columns
            else:
                raise TypeError("Invalid type {}".format(type(columns)))
        else:
            self.columns = columns
        self.drop = drop

    def fit(self, X, y=None, **fit_params):
        """Selecting skewed columns from the dataset.

        Parameters
        ----------
        X : {Dataframe}, shape = [n_samples, n_features]
            Dataframe, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        --------
        self
        """
        if self.columns is None:
            self.columns = X.select_dtypes(exclude=["object"]).columns

        if isinstance(X, pd.core.frame.DataFrame):
            try:
                _test = X[self.columns].astype(np.float32)
                del(_test)
            except ValueError:
                raise NameError("Null or categorical variables are not allowed: {}".format(X.dtypes))
        else:
            raise NameError("Invalid type {}".format(type(X)))

        skewness = X[self.columns].apply(lambda x: skew(x))
        skewness = skewness[abs(skewness) > 0.5]
        self.skew_features = skewness.index

        return self

    def transform(self, X):
        """Trransformer applies log to skewed features.

        Parameters
        ----------
        X : {Dataframe}, shape = [n_samples, n_features]
            Dataframe of samples, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        -------
        X_transform : {DAtaframe}, shape = [n_samples, n_features]
            A copy of the input Dataframe with the columns centered.

        """
        if not hasattr(self, "skew_features"):
            raise AttributeError("FixSkewness has not been fitted, yet.")

        if isinstance(X, pd.core.frame.DataFrame):
            X_transform = X.copy()
            X_transform[self.skew_features] = np.log1p(X[self.skew_features])
        else:
            raise NameError("Invalid type {}".format(type(X)))

        if self.drop:
            X_drop = [i for i in X.columns.tolist() if i not in self.columns]
            X_transform = X_transform.drop(X_drop, axis=1)

        return X_transform




"""Jaime Sendra Berenguer-2020.
MLearner Machine Learning Library Extensions
Author:Jaime Sendra Berenguer<www.linkedin.com/in/jaisenbe>
License: MIT
"""

import pandas as pd
import numpy as np
from sklearn.utils import check_array
from sklearn.preprocessing import LabelEncoder
from scipy import sparse
from sklearn.base import BaseEstimator, TransformerMixin
from ..externals.estimator_checks import check_is_fitted


class OneHotEncoder(BaseEstimator, TransformerMixin):

    """This transformer applies One-Hot-Encoder to features.

    Attributes
    ----------
    numerical: pandas [n_columns].
        numerical columns to be treated as categorical.
    columns:  pandas [n_columns].
        columns to use (if None then all categorical variables are included).

    Examples
    --------
    For usage examples, please see:
    https://jaisenbe58r.github.io/MLearner/user_guide/preprocessing/OneHotEncoder/


    """
    def __init__(self, columns=None, numerical=[], Drop=True):
        """Init OneHotEncoder."""
        if columns is not None:
            if isinstance(columns, list) or isinstance(columns, tuple):
                self.columns = columns
            else:
                raise TypeError("Invalid type {}".format(type(columns)))
        else:
            self.columns = columns

        if numerical != []:
            if isinstance(numerical, list) or isinstance(numerical, tuple):
                self.numerical = numerical
            else:
                raise TypeError("Invalid type {}".format(type(numerical)))
        else:
            self.numerical = numerical

        if isinstance(Drop, bool):
            self.Drop = Drop
        else:
            raise TypeError("Invalid type {}".format(type(Drop)))

    def fit(self, X, y=None, **fit_params):
        """Selecting OneHotEncoder columns from the dataset.

        Parameters
        ----------
        X : {Dataframe}, shape = [n_samples, n_features]
            Dataframe, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        --------
        self
        """
        if self.columns is None:
            self.columns = X.select_dtypes(include=["object"]).columns.tolist()

        if not isinstance(X, pd.core.frame.DataFrame) and not isinstance(X, np.ndarray):
            raise NameError("Invalid type {}".format(type(X)))

        self.columns += self.numerical
        # get all possible column values to filter not seen values
        self.allowed_columns = ["{}_{}".format(column, val) for column in self.columns for val in X[column].unique()]

        return self

    def transform(self, X):
        """Trransformer applies log to skewed features.

        Parameters
        ----------
        X : {Dataframe}, shape = [n_samples, n_features]
            Dataframe of samples, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        -------
        X_transform : {DAtaframe}, shape = [n_samples, n_features]
            A copy of the input Dataframe with the columns encoder.

        """
        check_is_fitted(self, 'allowed_columns')
        # if not hasattr(self, "allowed_columns"):
        #     raise AttributeError("FixSkewness has not been fitted, yet.")

        if isinstance(X, pd.core.frame.DataFrame):
            X_transform = X.copy()
        elif isinstance(X, np.ndarray):
            X_transform = X
        else:
            raise NameError("Invalid type {}".format(type(X)))

        # cast numerical columns to strings.
        for col in X_transform[self.columns].select_dtypes(exclude=["object"]).columns:
            X_transform[col] = X_transform[col].astype('str')

        one_hots = pd.get_dummies(X_transform[self.columns], prefix=self.columns)
        missing_cols = set(self.allowed_columns) - set(one_hots.columns)

        for c in missing_cols:
            one_hots[c] = 0

        if self.Drop:
            X_transform1 = X_transform.drop(self.columns, axis=1)
        else:
            X_transform1 = X_transform
        return pd.concat([X_transform1, one_hots.filter(self.allowed_columns).astype(np.int32)], axis=1)


class CategoricalEncoder(BaseEstimator, TransformerMixin):
    """Encode categorical features as a numeric array.
    The input to this transformer should be a matrix of integers or strings,
    denoting the values taken on by categorical (discrete) features.
    The features can be encoded using a one-hot aka one-of-K scheme
    (``encoding='onehot'``, the default) or converted to ordinal integers
    (``encoding='ordinal'``).
    This encoding is needed for feeding categorical data to many scikit-learn
    estimators, notably linear models and SVMs with the standard kernels.
    Read more in the :ref:`User Guide <preprocessing_categorical_features>`.

    Parameters
    ----------
    encoding : str, 'onehot', 'onehot-dense' or 'ordinal'
        The type of encoding to use (default is 'onehot'):
        - 'onehot': encode the features using a one-hot aka one-of-K scheme
          (or also called 'dummy' encoding). This creates a binary column for
          each category and returns a sparse matrix.
        - 'onehot-dense': the same as 'onehot' but returns a dense array
          instead of a sparse matrix.
        - 'ordinal': encode the features as ordinal integers. This results in
          a single column of integers (0 to n_categories - 1) per feature.
    categories : 'auto' or a list of lists/arrays of values.
        Categories (unique values) per feature:
        - 'auto' : Determine categories automatically from the training data.
        - list : ``categories[i]`` holds the categories expected in the ith
          column. The passed categories are sorted before encoding the data
          (used categories can be found in the ``categories_`` attribute).
    dtype : number type, default np.float64
        Desired dtype of output.
    handle_unknown : 'error' (default) or 'ignore'
        Whether to raise an error or ignore if a unknown categorical feature is
        present during transform (default is to raise). When this is parameter
        is set to 'ignore' and an unknown category is encountered during
        transform, the resulting one-hot encoded columns for this feature
        will be all zeros.
        Ignoring unknown categories is not supported for
        ``encoding='ordinal'``.

    Attributes
    ----------
    categories_ : list of arrays
        The categories of each feature determined during fitting. When
        categories were specified manually, this holds the sorted categories
        (in order corresponding with output of `transform`).

    Examples
    --------
    Given a dataset with three features and two samples, we let the encoder
    find the maximum value per feature and transform the data to a binary
    one-hot encoding.
    >>> from sklearn.preprocessing import CategoricalEncoder
    >>> enc = CategoricalEncoder(handle_unknown='ignore')
    >>> enc.fit([[0, 0, 3], [1, 1, 0], [0, 2, 1], [1, 0, 2]])
    ... # doctest: +ELLIPSIS
    CategoricalEncoder(categories='auto', dtype=<... 'numpy.float64'>,
              encoding='onehot', handle_unknown='ignore')
    >>> enc.transform([[0, 1, 1], [1, 0, 4]]).toarray()
    array([[ 1.,  0.,  0.,  1.,  0.,  0.,  1.,  0.,  0.],
           [ 0.,  1.,  1.,  0.,  0.,  0.,  0.,  0.,  0.]])

    See also
    --------
    sklearn.preprocessing.OneHotEncoder : performs a one-hot encoding of
      integer ordinal features. The ``OneHotEncoder assumes`` that input
      features take on values in the range ``[0, max(feature)]`` instead of
      using the unique values.
    sklearn.feature_extraction.DictVectorizer : performs a one-hot encoding of
      dictionary items (also handles string-valued features).
    sklearn.feature_extraction.FeatureHasher : performs an approximate one-hot
      encoding of dictionary items or strings.
    """

    def __init__(self, encoding='onehot', categories='auto', dtype=np.float64,
                 handle_unknown='error'):
        self.encoding = encoding
        self.categories = categories
        self.dtype = dtype
        self.handle_unknown = handle_unknown

    def fit(self, X, y=None):
        """Fit the CategoricalEncoder to X.
        Parameters
        ----------
        X : array-like, shape [n_samples, n_feature]
            The data to determine the categories of each feature.
        Returns
        -------
        self
        """

        if self.encoding not in ['onehot', 'onehot-dense', 'ordinal']:
            template = ("encoding should be either 'onehot', 'onehot-dense' "
                        "or 'ordinal', got %s")
            raise ValueError(template % self.handle_unknown)

        if self.handle_unknown not in ['error', 'ignore']:
            template = ("handle_unknown should be either 'error' or "
                        "'ignore', got %s")
            raise ValueError(template % self.handle_unknown)

        if self.encoding == 'ordinal' and self.handle_unknown == 'ignore':
            raise ValueError("handle_unknown='ignore' is not supported for"
                             " encoding='ordinal'")

        X = check_array(X, dtype=np.object, accept_sparse='csc', copy=True)
        _, n_features = X.shape

        self._label_encoders_ = [LabelEncoder() for _ in range(n_features)]

        for i in range(n_features):
            le = self._label_encoders_[i]
            Xi = X[:, i]
            if self.categories == 'auto':
                le.fit(Xi)
            else:
                valid_mask = np.in1d(Xi, self.categories[i])
                if not np.all(valid_mask):
                    if self.handle_unknown == 'error':
                        diff = np.unique(Xi[~valid_mask])
                        msg = ("Found unknown categories {0} in column {1}"
                               " during fit".format(diff, i))
                        raise ValueError(msg)
                le.classes_ = np.array(np.sort(self.categories[i]))

        self.categories_ = [le.classes_ for le in self._label_encoders_]

        return self

    def transform(self, X):
        """Transform X using one-hot encoding.

        Parameters
        ----------
        X : array-like, shape [n_samples, n_features]
            The data to encode.

        Returns
        -------
        X_out : sparse matrix or a 2-d array
            Transformed input.
        """
        X = check_array(X, accept_sparse='csc', dtype=np.object, copy=True)
        n_samples, n_features = X.shape
        X_int = np.zeros_like(X, dtype=np.int)
        X_mask = np.ones_like(X, dtype=np.bool)

        for i in range(n_features):
            valid_mask = np.in1d(X[:, i], self.categories_[i])

            if not np.all(valid_mask):
                if self.handle_unknown == 'error':
                    diff = np.unique(X[~valid_mask, i])
                    msg = ("Found unknown categories {0} in column {1}"
                           " during transform".format(diff, i))
                    raise ValueError(msg)
                else:
                    # Set the problematic rows to an acceptable value and
                    # continue `The rows are marked `X_mask` and will be
                    # removed later.
                    X_mask[:, i] = valid_mask
                    X[:, i][~valid_mask] = self.categories_[i][0]
            X_int[:, i] = self._label_encoders_[i].transform(X[:, i])

        if self.encoding == 'ordinal':
            return X_int.astype(self.dtype, copy=False)

        mask = X_mask.ravel()
        n_values = [cats.shape[0] for cats in self.categories_]
        n_values = np.array([0] + n_values)
        indices = np.cumsum(n_values)

        column_indices = (X_int + indices[:-1]).ravel()[mask]
        row_indices = np.repeat(np.arange(n_samples, dtype=np.int32),
                                n_features)[mask]
        data = np.ones(n_samples * n_features)[mask]

        out = sparse.csc_matrix((data, (row_indices, column_indices)),
                                shape=(n_samples, indices[-1]),
                                dtype=self.dtype).tocsr()
        if self.encoding == 'onehot-dense':
            return out.toarray()
        else:
            return out



"""Jaime Sendra Berenguer-2020.
MLearner Machine Learning Library Extensions
Author:Jaime Sendra Berenguer<www.linkedin.com/in/jaisenbe>
License: MIT
"""

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from sklearn.base import BaseEstimator, TransformerMixin
from ..externals.estimator_checks import check_is_fitted
# https://www.kaggle.com/funkyfrankie/sklearn-pipelines-and-transformers


class FeatureDropper(BaseEstimator, TransformerMixin):

    """Column drop according to the selected feature.

    Attributes
    ----------
    drop: list of features to drop [n_columns]

    Examples
    --------
    For usage examples, please see
    https://jaisenbe58r.github.io/MLearner/user_guide/preprocessing/FeatureDropper/

    """

    def __init__(self, drop=[]):
        """Init Feature droped."""
        if isinstance(drop, list):
            self.drop = drop
        else:
            raise NameError("Invalid type {}".format(type(drop)))

    def fit(self, X, y=None, **fit_params):
        """Gets the columns that not drop.

        Parameters
        ----------
        X : {Dataframe}, shape = [n_samples, n_features]
            Dataframe, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        --------
        self

        """
        if isinstance(X, pd.core.frame.DataFrame):
            self.col_drop = self.drop
        else:
            raise NameError("Invalid type {}".format(type(X)))

        return self

    def transform(self, X, **fit_params):
        """Features drop.

        Parameters
        ----------
        X : {Dataframe}, shape = [n_samples, n_features]
            Dataframe of samples, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        -------
        X_transform : {Dataframe}, shape = [n_samples, n_features]
            A copy of the input Dataframe with the columns dropped.

        """
        if not hasattr(self, "col_drop"):
            raise AttributeError("FeatureDropper has not been fitted, yet.")

        if not isinstance(X, pd.core.frame.DataFrame):
            raise NameError("Invalid type {}".format(type(X)))

        return X.drop(self.col_drop, axis=1)


class DropOutliers(BaseEstimator, TransformerMixin):

    """Drop Outliers from dataframe

    Attributes
    ----------
    features: `list`or `tuple
        list of features to drop outliers [n_columns]
    display: `boolean`
        Show histogram with changes made.

    Examples
    --------
    For usage examples, please see
    https://jaisenbe58r.github.io/MLearner/user_guide/preprocessing/DropOutliers/

    """

    def __init__(self, features=[], display=False):
        """Init Feature droped."""
        if isinstance(features, list) or isinstance(features, tuple):
            self.features = features
        else:
            raise TypeError("Invalid type {}".format(type(features)))

        if isinstance(display, bool):
            self.display = display
        else:
            raise TypeError("Invalid type {}".format(type(display)))

    def fit(self, X, y=None, **fit_params):
        """Gets the columns that not drop.

        Parameters
        ----------
        X : {Dataframe}, shape = [n_samples, n_features]
            Dataframe, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        --------
        self

        """
        if self.features == []:
            self.features = X.select_dtypes(exclude=["object"]).columns.tolist()

        if isinstance(X, pd.core.frame.DataFrame):
            try:
                _test = X[self.features].astype(np.float32)
                del(_test)
            except ValueError:
                raise ValueError("Null or categorical variables are not allowed: {}".format(X.dtypes))
        else:
            raise TypeError("Invalid type {}".format(type(X)))

        self._fit = True

        return self

    def transform(self, X, **fit_params):
        """Features drop.

        Parameters
        ----------
        X : {Dataframe}, shape = [n_samples, n_features]
            Dataframe of samples, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        -------
        X_transform : {Dataframe}, shape = [n_samples, n_features]
            A copy of the input Dataframe with the columns dropped.

        """
        check_is_fitted(self, '_fit')

        if not isinstance(X, pd.core.frame.DataFrame):
            raise TypeError("Invalid type {}".format(type(X)))

        if self.display:
            height, width = self._get_size_plot()
            figure, axs = plt.subplots(height, width, figsize=(width*3, height*3))
            ax = axs.flatten()

        data_transform = X.copy()
        for i in range(len(self.features)):
            _f = self.features[i]

            # Rango intercuartilico (Q3–Q1)
            IQR = data_transform[_f].quantile(0.75)-data_transform[_f].quantile(0.25)

            # se consideran atipicos los valores inferiores a Q1–1.5·IQR o superiores a Q3+1.5·IQR
            IQR_max = data_transform[_f].quantile(0.75) + 1.5*IQR
            IQR_min = data_transform[_f].quantile(0.25) - 1.5*IQR

            data_transform = data_transform[(data_transform[_f] > IQR_min) & (data_transform[_f] < IQR_max)]

            if self.display:
                ax[i].hist(data_transform[_f], density=True, color="g", alpha=0.7)
                ax[i].hist(X[_f], density=True, color="r", alpha=0.7)
                ax[i].set_title(_f)
                ax[i].legend(["Original", "Cleaned"])

        if self.display:
            figure.tight_layout()
            plt.show()

        return data_transform

    def _get_size_plot(self, width=4):
        """Size of plot."""
        check_is_fitted(self, '_fit')

        _n = len(self.features)

        if not _n <= width:
            width = 4
            nm = _n/width
        else:
            if _n > 1:
                width = _n
            else:
                width = 2
            nm = 0
        height = int(nm)+1

        return height, width



"""
Jaime Sendra Berenguer-2020.
MLearner Machine Learning Library Extensions
Author:Jaime Sendra Berenguer<www.linkedin.com/in/jaisenbe>
License: MIT
"""

import os
import pandas as pd
import numpy as np

import tensorflow as tf
import tensorflow_datasets as tfds

import re
from bs4 import BeautifulSoup


class Processor_data():
    """
    The DCNN class corresponds to the Neural Convolution Network algorithm
    for Natural Language Processing.

    Parameters
    ----------
        mame:
            Instance class name

    Attributes
    -----------
        clean: function
            Modulo limpieza de texto por medio de expresiones regulares


    Examples:
    ---------
        ```python
        cols = ["sentiment", "id", "date", "query", "user", "text"]
        data = pd.read_csv(
            TRAIN,
            header=None,
            names=cols,
            engine="python",
            encoding="latin1"
        )
        data.drop(["id", "date", "query", "user"],
                axis=1,
                inplace=True)
        nlptrans = Processor()
        data_process = nlptrans.process_text(data)
        ```
    """
    def __init__(
            self,
            target_vocab_size=2**16,
            language="en",
            value=0,
            padding='post',
            function=None,
            name="NLP"
            ):
        self.name = name
        self.target_vocab_size = target_vocab_size
        self.language = language
        self.value = value
        self.padding = padding
        self.function = function

    def clean(self, data):
        """
        Clean text.
        """
        if isinstance(data, np.ndarray):
            return [self._clean_text(_text[0]) for _text in data]

        elif isinstance(data, pd.core.frame.DataFrame) or isinstance(data, pd.core.series.Series):
            if data.isna().sum()[0] > 0:
                data = data.fillna(" ")
            return [self._clean_text(_text[0]) for _text in data.values]

        elif isinstance(data, list):
            return [self._clean_text(_text) for _text in data]

        else:
            data = data.split("\n")
            return [self._clean_text(_text) for _text in data]

    def apply_non_breaking_prefix(self, text, language="en"):
        """
        clean words with a period at the end to make it easier for us to use.

        Parameters
        ----------
            text:
                Text to apply cleaning.
            language: str
                Language a nonbreaking_prefix. options: en / es / fr.
        """
        this_dir, _ = os.path.split(__file__)
        _DATA_PATH = os.path.join(this_dir, "data", "Non-Breaking-Prefix")
        if language == "en":
            non_breaking_prefix = open_txt(os.path.join(_DATA_PATH, "P85-Non-Breaking-Prefix.en"))
        elif language == "es":
            non_breaking_prefix = open_txt(os.path.join(_DATA_PATH, "P85-Non-Breaking-Prefix.es"))
        elif language == "fr":
            non_breaking_prefix = open_txt(os.path.join(_DATA_PATH, "P85-Non-Breaking-Prefix.fr"))
        else:
            raise NameError(f"Language {language} not implemented")
        # Añadimos punto detras del prefijo
        non_breaking_prefix = non_breaking_prefix.split("\n")
        non_breaking_prefix = [' ' + pref + '.' for pref in non_breaking_prefix]
        for prefix in non_breaking_prefix:
            text = text.replace(prefix, prefix + '$$$')
        text = re.sub(r"\.(?=[0-9]|[a-z]|[A-Z])", ".$$$", text)
        text = re.sub(r"\.\$\$\$", '', text)
        return text

    def _clean_text(self, text, split=None, non_breaking_prefix=True):
        """
        Regular expressions applied to text.

        ```python
        def Function_clean(non_breaking_prefix_en):
            text = BeautifulSoup(text, "lxml").get_text()
            # Eliminamos la @ y su mención
            text = re.sub(r"@[A-Za-z0-9]+", ' ', text)
            # Eliminamos los links de las URLs
            text = re.sub(r"https?://[A-Za-z0-9./]+", ' ', text)
            # Nos quedamos solamente con los caracteres
            text = re.sub(r"[^a-zA-Z.!?']", ' ', text)
            return text
        ````
        """
        if str(text) == 'nan':
            text = " "
            print(text)

        def Function_clean(text):
            """
            Función por defecto
            """
            text = BeautifulSoup(text, "lxml").get_text()
            # Eliminamos la @ y su mención
            text = re.sub(r"@[A-Za-z0-9]+", ' ', text)
            # Eliminamos los links de las URLs
            text = re.sub(r"https?://[A-Za-z0-9./]+", ' ', text)
            # Nos quedamos solamente con los caracteres
            text = re.sub(r"[^a-zA-Z.!?']", ' ', text)
            return text

        if self.function is None:
            text = Function_clean(text)
        else:
            text = self.function(text)

        if non_breaking_prefix:
            text = self.apply_non_breaking_prefix(text)
        # Eliminamos espacios en blanco adicionales
        text = re.sub(r" +", ' ', text)
        # Se aplica split. >> p.e. text.split("\n")
        if split is not None:
            text = text.split(split)
        return text

    def _tokenizer(self, data):
        """
        Tokenizador:
        """
        self.tokenizer = tfds.features.text.SubwordTextEncoder.build_from_corpus(
            data, target_vocab_size=self.target_vocab_size
        )

    def encode_data(self, data, eval=False):
        """
        Encoder all text
        """
        if not eval:
            self._tokenizer(data=data)
        return [self.tokenizer.encode(sentence) for sentence in data]

    def apply_padding(self, data, eval=False):
        """
        El Pdding es una forma especial de enmascaramiento donde los pasos enmascarados
        se encuentran al comienzo o al comienzo de una secuencia. El padding proviene de
        la necesidad de codificar datos de secuencia en lotes contiguos: para que todas las
        secuencias en un lote se ajusten a una longitud estándar dada, es necesario rellenar
        o truncar algunas secuencias.
        """
        if not eval:
            self._MAX_LEN = max([len(sentence) for sentence in data])
        return tf.keras.preprocessing.sequence.pad_sequences(
            data,
            value=self.value,
            padding=self.padding,
            maxlen=self._MAX_LEN
            )

    def process_text(self, data, eval=False, isclean=False, padding=True):
        """
        Procesador completo de texto:
         - Limpieza con expresiones regulares
         - Tokenizador
         - Padding
        """
        # Cleaning
        if not isclean:
            data = self.clean(data)
        # List of strings
        else:
            if isinstance(data, np.ndarray):
                data = [_text[0] for _text in data]
            elif isinstance(data, pd.core.frame.DataFrame) or isinstance(data, pd.core.series.Series):
                if data.isna().sum()[0] > 0:
                    data = data.fillna(" ")
                data = [_text[0] for _text in data.values]
            elif isinstance(data, list):
                data = [_text for _text in data]
            else:
                data = data.split("\n")
                data = [_text for _text in data]
        # Encoding
        data = self.encode_data(data, eval=eval)
        # Padding
        if padding:
            return self.apply_padding(data, eval)
        else:
            return data


def delete_max_length(data_in, data_out=None, max_length=20):
    """
    Delete phrases with more than `max_length` characters.
    """
    idx_to_remove = [count for count, sent in enumerate(data_in)
                        if len(sent) > max_length]
    print(idx_to_remove)
    for idx in reversed(idx_to_remove):
        del data_in[idx]
        if data_out is not None:
            del data_out[idx]

    if data_out is not None:
        idx_to_remove = [count for count, sent in enumerate(data_out)
                            if len(sent) > max_length]
        for idx in reversed(idx_to_remove):
            del data_in[idx]
            del data_out[idx]
        return data_in, data_out
    else:
        return data_in


def open_txt(filename, encoding="utf-8"):
    """
    Function to open a .txt and return list of phrases.

    Parameters
    ----------
        filename:
            Path where the file is hosted.

        encoding:
            Unicode and text encodings.
    """
    with open(filename, mode="r", encoding=encoding) as f:
        text = f.read()
    return text




"""Jaime Sendra Berenguer-2020.
MLearner Machine Learning Library Extensions
Author:Jaime Sendra Berenguer<www.linkedin.com/in/jaisenbe>
License: MIT
"""

import pandas as pd
import numpy as np


def minmax_scaling(X, columns, min_val=0, max_val=1):

    """In max scaling of pandas DataFrames.

    Parameters
    --------

    array : pandas DataFrame, shape = [n_rows, n_columns].
    columns : array-like, shape = [n_columns]
        Array-like with column names, e.g., ['col1', 'col2', ...]
        or column indices [0, 2, 4, ...]
    min_val : `int` or `float`, optional (default=`0`)
        minimum value after rescaling.
    max_val : `int` or `float`, optional (default=`1`)
        maximum value after rescaling.

    Returns
    --------

    df_new : pandas DataFrame object.
        Copy of the array or DataFrame with rescaled columns.

    Examples
    --------

    For usage examples, please see
    http://jaisenbe58r.github.io/mlearner/user_guide/preprocessing/minmax_scaling/.


    adapted from
    https://github.com/rasbt/mlxtend/blob/master/mlxtend/preprocessing/scaling.py
    Author: Sebastian Raschka <sebastianraschka.com>
    License: BSD 3 clause

    """

    ary_new = X[columns].astype(float)
    if len(ary_new.shape) == 1:
        ary_new = ary_new[:, np.newaxis]

    if isinstance(ary_new, pd.DataFrame):
        ary_newt = ary_new.loc
    else:
        raise AttributeError('Input array must be a pandas')

    numerator = ary_newt[:, columns] - ary_newt[:, columns].min(axis=0)
    denominator = (ary_newt[:, columns].max(axis=0) -
                   ary_newt[:, columns].min(axis=0))
    ary_newt[:, columns] = numerator / denominator

    if not min_val == 0 and not max_val == 1:
        ary_newt[:, columns] = (ary_newt[:, columns] *
                                (max_val - min_val) + min_val)
    return ary_newt[:, columns]




"""Jaime Sendra Berenguer-2020.
MLearner Machine Learning Library Extensions
Author:Jaime Sendra Berenguer<www.linkedin.com/in/jaisenbe>
License: MIT
"""

import os
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from mlearner.load import DataLoad


class DataExploratory(DataLoad):

    """Class to preprocessed object for data cleaning.

    Attributes
    ----------
    data: `pd.DataFrame` of Dataset

    Examples
    --------
    For usage examples, please see
    https://jaisenbe58r.github.io/MLearner/user_guide/preprocessing/DataCleaner/

    """
    def __init__(self, data):
        super().__init__(data)
        """
        Inicializacion de la clase de Preprocesado de un dataframe
        """
        if isinstance(data, pd.core.frame.DataFrame):
            self.data = data
        else:
            raise TypeError("Invalid type {}".format(type(data)))

    def dtypes(self, X=None):
        """
        retorno del tipo de datos por columna
        """
        if X is None:
            return self.data.dtypes
        else:
            return X.dtypes

    def missing_values(self, X=None):
        """
        Numero de valores vacios en el dataframe.
        """
        if X is None:
            # Number of missing in each column
            missing = pd.DataFrame(self.data.isnull().sum()).rename(columns={0: 'total'})
        else:
            missing = pd.DataFrame(X.isnull().sum()).rename(columns={0: 'total'})

        _miss = missing.sort_values('total', ascending=False)
        return _miss

    def isNull(self):
        if not self.missing_values()["total"].values.sum() == 0:
            print("Cuidado que existen valores nulos")
            return True
        else:
            print("No existen valores nulos")
            return False

    def view_features(self):
        """
        Mostrar features del dataframe
        """
        return list(self.data.columns)

    def categorical_vs_numerical(self):
        categorical_list = []
        numerical_list = []
        for i in self.data.columns.tolist():
            if self.data[i].dtype == 'object':
                categorical_list.append(i)
            else:
                numerical_list.append(i)
        print('Number of categorical features:', str(len(categorical_list)))
        print('Number of numerical features:', str(len(numerical_list)))

        return categorical_list, numerical_list

    def type_object(self):
        """
        Deteccion de de categorias con type "object"
        """
        var = [i for i in self.view_features() if self.data[i].dtype == np.object]
        return var

    def not_type_object(self):
        """
        Deteccion de de categorias con type "object" 
        """
        var = [i for i in self.view_features() if not self.data[i].dtype == np.object]
        return var


class DataAnalyst(DataExploratory):

    """Class for Preprocessed object for data analysis.

    Attributes
    ----------
    data: pd.DataFrame of Dataset

    Examples
    --------
    For usage examples, please see
    https://jaisenbe58r.github.io/MLearner/user_guide/preprocessing/DataAnalyst/

    """
    def __init__(self, data):
        super().__init__(data)
        """
        Inicializacion de la clase de Analisis de datos
        """
        if isinstance(data, pd.core.frame.DataFrame):
            self.data = data
        else:
            raise TypeError("Invalid type {}".format(type(data)))
        self.data = data

    def _get_size_plot(self, features, width=4):
        """Size of plot."""
        _n = len(features)
        if not _n <= width:
            nm = _n/width
        else:
            if _n > 1:
                width = _n
            else:
                width = 2
            nm = 0
        height = int(nm)+1

        return height, width

    def _check_parameters_features(self, features):

        if features is not None:
            if isinstance(features, list) or isinstance(features, tuple):
                if len(features) == 0:
                    raise NameError("Empty category List")
                else:
                    for i in features:
                        if i not in self.data.columns.tolist():
                            raise NameError("Feature '{}' not included in dataset targets".format(i))
                        else:
                            for i in self.data[features].dtypes:
                                if i == "object":
                                    raise NameError("Type Object not permited")
                            else:
                                _features = features
            else:
                raise TypeError("Invalid type {}".format(type(features)))
        else:
            _features = self.data.select_dtypes(exclude=["object"]).columns.tolist()

        return _features

    def _check_parameters_target(self, target):

        if target is not None:
            if isinstance(target, list) or isinstance(target, tuple):
                if len(target) == 0:
                    raise NameError("Empty category List")
                elif len(target) > 1:
                    raise NameError("Only one category column can be selected")
                else:
                    if target[0] not in self.data.columns.tolist():
                        raise NameError("Target '{}' not included in dataset columns".format(target))
                    else:
                        _target = target[0]
            else:
                raise TypeError("Invalid type {}".format(type(target)))
        else:
            raise NameError("Target can not be 'None'")

        return _target

    def _check_parameters_category(self, category, target):

        if category is not None:
            if isinstance(category, list) or isinstance(target, tuple):
                if len(category) == 0:
                    raise NameError("Empty category List")
                elif len(category) > 1:
                    raise NameError("Only one category can be selected")
                else:
                    if category[0] not in list(self.data[target].unique()):
                        raise NameError("Category '{}' not included in targets".format(category[0]))
                    else:
                        _category = category[0]
            else:
                raise TypeError("Invalid type {}".format(type(category)))
        else:
            raise NameError("category can not be 'None'")

        return _category

    def _check_path(self, path):
        if not os.path.isdir(path):
            raise NameError("Invalid path {}".format(path))

    def boxplot(self, features=None, target=None, display=False, save_image=False, path="/", width=2):
        """
        Funcion que realiza un BoxPlot sobre la dispesion de cada categoria
        respecto a los grupos de target.

        Inputs:
            - data: Datos generales del dataset.
            - features: categorias a analizar.
        """
        _features = self._check_parameters_features(features)
        _target = self._check_parameters_target(target)

        if save_image:
            self._check_path(path)

        _vars = [i for i in _features if not i == _target]

        height, width = self._get_size_plot(features=_features, width=width)
        figure, axs = plt.subplots(height, width, figsize=(width*5, height*3))
        ax = axs.flatten()

        datos = []
        for i in range(len(_vars)):
            datos = []
            _cont = list(self.data[_target].unique())
            for j in _cont:
                data_train_group_j = self.data.groupby(_target).get_group(j)
                datos.append(data_train_group_j[_vars[i]])
            ax[i].boxplot(datos, positions=np.array(range(2)))
            ax[i].set_title(_vars[i])
            ax[i].legend(list(self.data[_target].unique()))

        if display:
            figure.tight_layout()
            plt.show()

        if save_image:
            plt.savefig(path)

    def dispersion_categoria(self, features=None, target=None, density=True, display=False, width=2, save_image=False, path="/"):
        """
        Funcion que realiza un plot sobre la dispesion de cada categoria respecto a los grupos de target.

        Inputs:
            - data: Datos generales del dataset.
            - features: categorias a analizar.

        """
        _features = self._check_parameters_features(features)
        _target = self._check_parameters_target(target)

        if save_image:
            self._check_path(path)

        _vars = [i for i in _features if not i == _target]

        height, width = self._get_size_plot(features=_features, width=width)
        figure, axs = plt.subplots(height, width, figsize=(width*5, height*3))
        ax = axs.flatten()

        for i in range(len(_vars)):
            for j in list(self.data[_target].unique()):

                data_train_group_j = self.data.groupby(_target).get_group(j)
                ax[i].hist(data_train_group_j[_vars[i]], alpha=0.7, density=density)

            ax[i].set_title(_vars[i])
            ax[i].legend(list(self.data[_target].unique()))

        if display:
            figure.tight_layout()
            plt.show()

        if save_image:
            plt.savefig(path)

    def sns_jointplot(self, feature1, feature2, target=None, categoria1=None,
                        categoria2=None, display=True, save_image=False, path="/"):
        import seaborn as sns

        _feature1 = self._check_parameters_features(feature1)
        _feature2 = self._check_parameters_features(feature2)
        _target = self._check_parameters_target(target)

        _f = list(self.data[_target].unique())

        _categoria1 = self._check_parameters_category(categoria1, _target)

        elem = self.data.groupby(target).get_group(_categoria1)
        g = sns.jointplot(_feature1, _feature2, data=elem, kind="reg", truncate=False, color="m", height=7)

        if categoria2 is not None:
            _categoria2 = self._check_parameters_category(categoria2, _target)

            elem = self.data.groupby(target).get_group(_categoria2)
            g.x = elem[feature1]
            g.y = elem[feature2]
            g.plot_joint(plt.scatter, marker='x', c='b', s=50)

        if display:
            plt.show()

        if save_image:
            self._check_path(path)
            g.savefig(path)

    def sns_pairplot(self, features=None, target=None, display=True, save_image=False, path="/", palette="husl"):
        import seaborn as sns

        _features = self._check_parameters_features(features)
        _target = self._check_parameters_target(target)

        if _target not in _features:
            _features.append(_target)

        g = sns.pairplot(self.data[_features], hue=_target, palette=palette)

        if display:
            plt.show()

        if save_image:
            self._check_path(path)
            g.savefig(path)

    def distribution_targets(self, target=None, display=True, save_image=False, path="/", palette="Set2"):
        import seaborn as sns

        _target = self._check_parameters_target(target)

        total = self.data[_target].shape[0]
        ax = sns.countplot(x=_target, palette=palette, data=self.data)
        sns.set(font_scale=1.5)
        ax.set_xlabel(' ')
        ax.set_ylabel(' ')
        _ = plt.gcf()
        # fig.set_size_inches(10,5)
        ax.set_ylim(top=total)

        for p in ax.patches:
            ax.annotate('{:.2f}%'.format(100*p.get_height()/total), (p.get_x()+0.3, p.get_height()+50))

        plt.title('Distribution of {} Targets'.format(total))
        plt.xlabel('Categories')
        plt.ylabel('Frequency [%]')

        if display:
            plt.show()

        if save_image:
            self._check_path(path)
            plt.savefig(path)

    def corr_matrix(self, features=None, display=True, save_image=False, path="/"):
        """
        matriz de covarianza:

        Un valor positivo para r indica una asociacion positiva
        Un valor negativo para r indica una asociacion negativa.

        Cuanto mas cerca estar de 1cuanto mas se acercan los puntos de datos a una linea recta,
        la asociacion lineal es mas fuerte. Cuanto mas cerca este r de 0, lo que debilita la asociacion lineal.

        """
        import seaborn as sns

        _features = self._check_parameters_features(features)
        X = self.data[_features]

        sns.set(style="white")
        corr = X.corr()
        # Set up the matplotlib figure
        _, _ = plt.subplots(figsize=(11, 9))

        # Generate a custom diverging colormap
        cmap = sns.diverging_palette(220, 10, as_cmap=True)

        # Draw the heatmap with the mask and correct aspect ratio
        sns.heatmap(corr, cmap=cmap, vmax=.3, center=0,
                    square=True, linewidths=.5, cbar_kws={"shrink": .5})
        if display:
            plt.show()

        if save_image:
            self._check_path(path)
            plt.savefig(path)

    def Xy_dataset(self, target=None):
        """
        Separar datos del target en conjunto (X, y)
        """
        _target = self._check_parameters_target(target)
        y = _target
        x = [i for i in self.data.columns.tolist() if i not in y]

        self.X = self.data[x]
        self.y = self.data[y]




"""Jaime Sendra Berenguer-2020.
MLearner Machine Learning Library Extensions
Author:Jaime Sendra Berenguer<www.linkedin.com/in/jaisenbe>
License: MIT
"""

import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin
from ..externals.estimator_checks import check_is_fitted


class ExtractCategories(BaseEstimator, TransformerMixin):

    """This transformer filters the selected dataset categories.

    Attributes
    ----------
    categories: `list` of categories that you want to keep.

    Examples
    --------
    For usage examples, please see
    https://jaisenbe58r.github.io/MLearner/user_guide/preprocessing/ReplaceTransformer/

    """
    def __init__(self, categories=None, target=None):
        """Init filter categories"""
        if categories is not None:
            if isinstance(categories, list) or isinstance(categories, tuple):
                if len(categories) == 0:
                    raise NameError("Empty category List")
                else:
                    self.categories = categories
            else:
                raise TypeError("Invalid type {}".format(type(categories)))
        else:
            self.categories = categories

        if target is not None:
            if isinstance(target, list) or isinstance(target, tuple):
                if len(target) == 0:
                    raise NameError("Empty category List")
                elif len(target) > 1:
                    raise NameError("Only one category column can be selected")
                else:
                    self.target = target[0]
            else:
                raise TypeError("Invalid type {}".format(type(target)))
        else:
            self.target = target

    def fit(self, X, y=None, **fit_params):
        """Gets the columns to make filters the selected dataset categories.

        Parameters
        ----------
        X : {Dataframe}, shape = [n_samples, n_features]
            Dataframe, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        --------
        self

        """
        if not isinstance(X, pd.core.frame.DataFrame) and not isinstance(X, pd.core.series.Series):
            raise TypeError("Invalid type {}".format(type(X)))

        if self.target is None:
            if isinstance(X, pd.core.series.Series):
                raise NameError("especificar target distinto de None")
            else:
                self.target = X.columns.tolist()[-1]
        else:
            if isinstance(X, pd.core.series.Series):
                pass
            else:
                if self.target not in X.columns.tolist():
                    print(X.columns.tolist())
                    raise NameError("Target '{}' not included in dataset columns".format(self.target))

        if self.categories is None:
            if isinstance(X, pd.core.series.Series):
                self.categories = X.unique().tolist()
            else:
                self.categories = X[self.target].unique().tolist()
        else:
            for i in self.categories:
                if isinstance(X, pd.core.series.Series):
                    if i not in X.unique().tolist():
                        raise NameError("Category '{}' not included in dataset targets".format(i))
                else:
                    if i not in X[self.target].unique().tolist():
                        raise NameError("Category '{}' not included in dataset targets".format(i))

        self._fitted = True
        return self

    def transform(self, X):
        """Gets the columns to make filters the selected dataset categories.

        Parameters
        ----------
        X : {Dataframe}, shape = [n_samples, n_features]
            Dataframe of samples, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        -------
        X_transform : {Dataframe}, shape = [n_samples, n_features]
            A copy of the input Dataframe with the columns replaced.

        """
        check_is_fitted(self, '_fitted')

        if not isinstance(X, pd.core.frame.DataFrame) and not isinstance(X, pd.core.series.Series):
            raise TypeError("Invalid type {}".format(type(X)))

        X_transform = X.copy()
        if isinstance(X, pd.core.series.Series):
            X_transform = X[X == self.categories[0]]
        else:
            X_transform = X[X[self.target] == self.categories[0]]
        if len(self.categories) > 1:
            for i in range(len(self.categories)-1):
                if isinstance(X, pd.core.series.Series):
                    X_transform_aux = X[X == self.categories[i+1]]
                    X_transform = X_transform.append(X_transform_aux)
                else:
                    X_transform_aux = X[X[self.target] == self.categories[i+1]]
                    X_transform = pd.merge(left=X_transform, right=X_transform_aux, how="outer")

        return X_transform





"""Jaime Sendra Berenguer-2020.
MLearner Machine Learning Library Extensions
Author:Jaime Sendra Berenguer<www.linkedin.com/in/jaisenbe>
License: MIT
"""

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import os
import datetime
import joblib
import time

from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.model_selection import train_test_split
from sklearn.metrics import accuracy_score

import lightgbm as lgb
from lightgbm import LGBMClassifier

import seaborn as sns

from mlearner.training import Training
from mlearner.utils import ParamsManager

import warnings
warnings.filterwarnings("ignore")

param_file = "mlearner/classifier/config/models.json"


class modelLightBoost(Training, BaseEstimator, ClassifierMixin):
    """
    Ejemplo multiclass:
    https://www.kaggle.com/nicapotato/multi-class-lgbm-cv-and-seed-diversification
    """
    def __init__(self, name="LGB", random_state=99, train_dir="", params=None, *args, **kwargs):

        self.name = name
        self.train_dir = train_dir + "/" + "model_" + str(self.name) + "/"
        self.random_state = random_state

        if params is None:
            self.get_params_json()
            self.params.update({
                        'model_dir': self.train_dir,
                        "seed": self.random_state})
        else:
            # if isinstance(params)
            self.params = params

        self.model = LGBMClassifier(**self.params)
        super().__init__(self.model, random_state=self.random_state)

    def get_params_json(self):
        self.manager_models = ParamsManager(param_file, key_read="Models")
        self.params = self.manager_models.get_params()["LightBoost"]

        self.manager_finetune = ParamsManager(param_file, key_read="FineTune")
        self.params_finetune = self.manager_finetune.get_params()["LightBoost"]

    def dataset(self, X, y, categorical_columns_indices=None, test_size=0.2, *args, **kwarg):

        self.categorical_columns_indices = categorical_columns_indices
        self.X = X
        self.columns = list(X)

        self.y, self.cat_replace = self.replace_multiclass(y)

        self.X_train, self.X_test, self.y_train, self.y_test = train_test_split(self.X, self.y, test_size=test_size, random_state=self.random_state)

        self.dtrain = lgb.Dataset(self.X_train.values, label=self.y_train.values,
                                    feature_name=self.X_train.columns.tolist())
        self.dvalid = lgb.Dataset(self.X_test.values, label=self.y_test.values,
                                    feature_name=self.X_test.columns.tolist())
        self.all_train_data = lgb.Dataset(self.X.values, label=self.y.values,
                                            feature_name=self.X.columns.tolist())

    def set_dataset_nosplit(self, X_train, X_test, y_train, y_test, categorical_columns_indices=None, *args, **kwarg):

        self.categorical_columns_indices = categorical_columns_indices
        self.columns = list(X_train)
        _ytrain, _ = self.replace_multiclass(y_train)
        _ytest, _ = self.replace_multiclass(y_test)

        self.X_train = X_train
        self.X_test = X_test
        self.y_train = y_train
        self.y_test = y_test

        self.X = pd.concat([X_train, X_test], axis=0)
        self.y = pd.concat([y_train, y_test], axis=0)

        self.dtrain = lgb.Dataset(self.X_train.values, label=self.y_train.values,
                                    feature_name=self.X_train.columns.tolist())
        self.dvalid = lgb.Dataset(self.X_test.values, label=self.y_test.values,
                                    feature_name=self.X_test.columns.tolist())
        self.all_train_data = lgb.Dataset(self.X.values, label=self.y.values,
                                            feature_name=self.X.columns.tolist())

    def replace_multiclass(self, targets):

        _unic = targets.unique().tolist()
        _remp = np.arange(0, len(_unic)).tolist()
        return targets.replace(_unic, _remp), _unic

    def fit(self, X=None, y=None, X_train=None, X_test=None, y_train=None, y_test=None, mute=False,
            use_best_model=True, verbose=0, num_boost_round=100, nosplit=False, **kwargs):

        if not nosplit:
            self.dataset(X, y)
        else:
            self.set_dataset_nosplit(X_train, X_test, y_train, y_test)

        self.params.update({'verbose': verbose})

        self.model = lgb.train(
                                self.params,
                                self.dtrain,
                                num_boost_round=num_boost_round,
                                verbose_eval=verbose,
                                **kwargs)

        preds_test = [np.argmax(line) for line in self.model.predict(self.X_test, num_iteration=self.model.best_iteration)]
        score_test = accuracy_score(self.y_test, preds_test)

        preds_train = [np.argmax(line) for line in self.model.predict(self.X_train, num_iteration=self.model.best_iteration)]
        score_train = accuracy_score(self.y_train, preds_train)

        if not mute:
            print("Accurancy para el conjunto de entrenamiento ---> {:.2f}%".format(score_train*100))
            print("Accurancy para el conjunto de validacion ------> {:.2f}%".format(score_test*100))

    def fit_cv(self, X=None, y=None, X_train=None, X_test=None, y_train=None, y_test=None, nfold=5,
                use_best_model=True, verbose=200, nosplit=False, early_stopping_rounds=150, num_boost_round=2000, **kwargs):

        if not nosplit:
            self.dataset(X, y)
        else:
            self.set_dataset_nosplit(X_train, X_test, y_train, y_test)

        self.params.update({'verbose': verbose})
        self.lgb_cv = lgb.cv(
                                params=self.params,
                                train_set=self.all_train_data,
                                num_boost_round=num_boost_round,
                                stratified=True,
                                nfold=nfold,
                                seed=self.random_state,
                                early_stopping_rounds=early_stopping_rounds,
                                **kwargs)
        loss = self.params["metric"]
        optimal_rounds = np.argmin(self.lgb_cv[str(loss) + '-mean'])
        best_cv_score = min(self.lgb_cv[str(loss) + '-mean'])

        if not verbose == 0:
            print("\nOptimal Round: {}\nOptimal Score: {:.3f} + stdv:{:.3f}".format(
                optimal_rounds, best_cv_score, self.lgb_cv[str(loss) + '-stdv'][optimal_rounds]))

        results = {"Rounds": optimal_rounds,
                                "Score": best_cv_score,
                                "STDV": self.lgb_cv[str(loss) + '-stdv'][optimal_rounds],
                                "LB": None,
                                "Parameters": self.params}
        score = np.mean(self.lgb_cv[str(loss) + '-mean'])

        return score, results

    def func_acc(self, prob_pred, y_target):

        _y_pred = np.zeros(len(prob_pred))

        for i in range(0, len(prob_pred)):
            _y_pred[i] = int(np.argmax(prob_pred[i]))

        accuracy = accuracy_score(_y_pred, y_target)
        return accuracy

    def pred_binary(self, X, *args, **kwargs):
        _X_copy = X.loc[:, self.columns].copy()
        preds = self.model.predict(_X_copy, *args, **kwargs)
        return np.where(preds > 0.5, 1, 0)

    def pred_multiclass(self, X, *args, **kwargs):
        _X_copy = X.loc[:, self.columns].copy()
        return [np.argmax(line) for line in self.model.predict(_X_copy, num_iteration=self.model.best_iteration)]

    def update_model(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self.model, k, v)

    def save_model(self, direct="./checkpoints", name="LGM_model", file_model=".txt"):

        if not os.path.isdir(direct):
            try:
                os.mkdir(direct)
                print("Directorio creado: " + direct)
            except OSError as e:
                raise NameError("Error al crear el directorio")

        current_time = datetime.datetime.now().strftime("%Y%m%d-%H%M%S")

        if file_model == ".txt":
            filename = direct + "/" + name + "_" + current_time + ".txt"
            self.model.save_model(filename)
        elif file_model == ".pkl":
            filename = direct + "/" + name + "_" + current_time + ".pkl"
            joblib.dump(self.model, filename)
        else:
            raise NameError("Type {} not permited".format(file_model))
        print("Modelo guardado en la ruta: " + filename)

    def load_model(self, direct="./checkpoints/LGM_model.txt", file_model=".txt"):

        if not os.path.isdir(direct):
            print("no existe el drectorio especificado")

        if file_model == ".txt":
            self.model = LGBMClassifier(model_file=direct)
        elif file_model == ".pkl":
            self.model = joblib.load(direct)
        else:
            raise NameError("Type {} not permited".format(file_model))
        print("Modelo cargado de la ruta: " + direct)

    def predict(self, X, *args, **kwargs):
        _X_copy = X.loc[:, self.columns].copy()
        return self.model.predict(_X_copy, *args, **kwargs)

    def predict_proba(self, X, *args, **kwargs):
        _X_copy = X.loc[:, self.columns].copy()
        return self.model.predict_proba(_X_copy, *args, **kwargs)

    def index_features(self, features):

        _index = []
        for i in features:
            _index.append(self.X.columns.get_loc(i))
        if _index == []:
            raise NameError("No coincide ninguna de las features introducidas")
        return _index

    def get_important_features(self, display=True, max_num_features=20):

        if display:
            lgb.plot_importance(self.model, max_num_features=max_num_features, figsize=(6, 6), title='Feature importance (LightGBM)')
            plt.show()
        # return _feature_importance_df

    def FineTune_SearchCV(self, X=None, y=None, X_train=None, X_test=None, y_train=None, y_test=None, params=None, params_finetune=None, ROC=False,
                            randomized=True, cv=10, display_ROC=True, verbose=0, n_iter=10, replace_model=True, nosplit=False, finetune_dir=""):

        self.get_params_json()

        self.finetune_dir = finetune_dir + "/" + "model_finetune_" + str(self.name) + "/"
        self.params.update({
                    'train_dir': self.finetune_dir,
                    "seed": self.random_state})
        if params is not None:
            self.params = params
        if params_finetune is not None:
            self.params_finetune = params_finetune

        if not nosplit:
            self.dataset(X, y)
        else:
            self.set_dataset_nosplit(X_train, X_test, y_train, y_test)

        self.params.update({'verbosity': verbose})
        self.model = LGBMClassifier(**self.params)

        self._best_Parameters, self.results_df = self.FineTune(self.model, self.X_train, self.y_train, self.params_finetune,
                                                                cv=cv, randomized=True, n_iter=n_iter, verbose=1)
        self.params.update(**self._best_Parameters)
        self.fit(self.X_train, self.y_train)

        print("\n")
        score = accuracy_score(self.y_test, self.pred_multiclass(self.X_test))
        print("\n")
        print("Resultado del conjunto de test con los parametros optimos: {:.2f}%".format(score*100))
        print("\n")
        print("Report clasificacion con el conjunto de test: ")
        self.evaluate(self.model, self.X_test, self.y_test)
        print("\n")
        print("Validacion cruzada con todos los datos del dataset: ")
        print("\n")
        self.KFold_CrossValidation(LGBMClassifier(**self._best_Parameters), self.X, self.y, n_splits=cv, ROC=ROC, shuffle=True, mute=False,
                                    logdir_report="", display=True, save_image=True, verbose=0)
        return self._best_Parameters, self.results_df

    def SeedDiversification_cv(self, X=None, y=None, X_train=None, X_test=None, y_train=None, y_test=None, n_iter=10, n_max=2018-2022, cv=10,
                                nosplit=False, finetuneseed_dir="", display=True, save_image=True, verbose=0):
        allmodelstart = time.time()
        self.get_params_json()

        self.finetune_dir = finetuneseed_dir + "/" + "model_finetune_seed" + str(self.name) + "/"
        self.params.update({'train_dir': self.finetune_dir,
                            'verbosity': verbose})
        if not nosplit:
            self.dataset(X, y)
        else:
            self.set_dataset_nosplit(X_train, X_test, y_train, y_test)

        self.params.update({'verbosity': verbose})
        self.model = LGBMClassifier(**self.params)

        _rd = np.random.uniform(0, n_max, n_iter).astype(np.int32).tolist()
        params_finetuneseed = {"seed": _rd}
        del(_rd)

        self._best_Parameters, self.results_df = self.FineTune(self.model, self.X, self.y,
                                                                params_finetuneseed, randomized=False, cv=cv, n_iter=n_iter,
                                                                verbose=1, mute=True)
        print("All Model Runtime: %0.2f Minutes" % ((time.time() - allmodelstart)/60))

        print("Diversificacion de la semilla - mean AUC: {:.2f}% - std AUC: {:.5f}".format(
                    self.results_df['mean_test_AUC'].mean()*100, self.results_df['std_test_AUC'].mean()))

        print("Diversificacion de la semilla - mean Acc: {:.2f}% - std Acc: {:.5f}".format(
                    self.results_df['mean_test_Accuracy'].mean()*100, self.results_df['std_test_Accuracy'].mean()))
        return self._best_Parameters, self.results_df

    def SeedDiversification_fs(self, X, y, params, n_iter=10, mute=False, logdir_report="", display=True, save_image=True):

        allmodelstart = time.time()
        # Run Model with different Seeds
        all_feature_importance_df = pd.DataFrame()
        _y, _ = self.replace_multiclass(y)
        all_seeds = np.random.uniform(1, 1000, n_iter).astype(np.int32).tolist()

        for seeds_x in all_seeds:
            modelstart = time.time()
            print("Seed: ", seeds_x,)
            # Go Go Go
            params["seed"] = seeds_x
            model = lgb.train(
                params,
                lgb.Dataset(X.values, label=_y.values),
                verbose_eval=100)

            # Feature Importance
            fold_importance_df = pd.DataFrame()
            fold_importance_df["feature"] = X.columns.tolist()
            fold_importance_df["importance"] = model.feature_importance()
            all_feature_importance_df = pd.concat([all_feature_importance_df, fold_importance_df], axis=0)

            # Submit Model Individually
        #     seed_submit(model= lgb_final, seed= seeds_x, X_test)
            if not mute:
                print("Model Runtime: %0.2f seconds"%((time.time() - modelstart)))
                print("#"*50)
            del model

        cols = all_feature_importance_df[["feature", "importance"]].groupby("feature").mean().sort_values(by="importance", ascending=False)[:50].index
        best_features = all_feature_importance_df.loc[all_feature_importance_df.feature.isin(cols)]
        plt.figure(figsize=(8, 10))
        sns.barplot(x="importance", y="feature",
                    data=best_features.sort_values(by="importance", ascending=False))
        plt.title('LightGBM Features (avg over folds)')
        plt.tight_layout()
        if display:
            plt.show()
        if save_image:
            filename = logdir_report + 'lgb_importances.png'
            plt.savefig(filename)
        print("All Model Runtime: %0.2f Minutes" % ((time.time() - allmodelstart)/60))

    def __getattr__(self, attr):
        """
        Pass all other method calls to self.model.
        """
        return getattr(self.model, attr)



"""
Jaime Sendra Berenguer-2020.
MLearner Machine Learning Library Extensions
Author:Jaime Sendra Berenguer<www.linkedin.com/in/jaisenbe>
License: MIT
"""

from .cnn import DCNN
from .utils import Processor_data, open_txt
from mlearner.nlp import helpers
from .cnn_advanced import DCNN_Advanced
from .transformer import Transformer, Transformer_train

__all__ = ["DCNN", "Processor_data", "open_txt", "helpers", "DCNN_Advanced",
            "Transformer", "Transformer_train"]




"""Jaime Sendra Berenguer-2020.
MLearner Machine Learning Library Extensions
Author:Jaime Sendra Berenguer <www.linkedin.com/in/jaisenbe>
License: MIT
"""


# from .voting import MajorityVoteClassifier


# __all__ = ["MajorityVoteClassifier"]



"""
Jaime Sendra Berenguer-2020.
MLearner Machine Learning Library Extensions
Author:Jaime Sendra Berenguer<www.linkedin.com/in/jaisenbe>
License: MIT
"""

import numpy as np
import math

import tensorflow as tf
from tensorflow.keras import layers

"""
Based on https://www.udemy.com/course/procesamiento-del-lenguaje-natural/learn/lecture/21502260#overview

Author: Juan Gabriel Gomila
Course: "Procesamiento del Lenguaje Natural Moderno en Python.

Original Text: https://arxiv.org/pdf/1404.2188.pdf

"""

"""
Build model layers.
"""


class Widening(layers.Layer):
    """
    Widening Layer: Padding for an extended 2D convolution

    input: (batches, seq_len, d_model, channels)

    Parameters
    ----------
        padding_size: int
            Padding size of the algorithm input text.

        padding_value: float
            Padding value.
    """
    def __init__(self, padding_size, padding_value=0., axis=1):
        super(Widening, self).__init__()
        self.padding_size = padding_size
        self.padding_value = padding_value
        self.axis = axis

    def build(self, input_shape):
        paddings_arr = np.array([[0, 0]]*len(input_shape))
        paddings_arr[self.axis, :] = [self.padding_size, self.padding_size]
        self.paddings = tf.convert_to_tensor(paddings_arr, dtype="int32")

    def call(self, inputs):
        return tf.pad(inputs,
                      self.paddings,
                      constant_values=self.padding_value)


class MyConv2D(layers.Layer):
    """
    MyConv2D Layer: 2D convolution

    Parameters
    ----------
        nb_filters: int
            Number of filters.

        conv_width: float
            Convolution 2D size

        padding: int
            Padding size of the algorithm input text.

        emb_dim: float
            Embeding size of the algorithm input text.
    """
    def __init__(self,
                 nb_filters,
                 conv_width,
                 padding,
                 emb_dim):
        super(MyConv2D, self).__init__()
        self.emb_dim = emb_dim
        self.conv_per_col = [layers.Conv2D(filters=nb_filters,
                                           kernel_size=[conv_width, 1],
                                           padding="valid",
                                           activation="tanh")
                             for _ in range(emb_dim)]

    def call(self, inputs):
        convolutions = []
        for i in range(self.emb_dim):
            convolutions.append(self.conv_per_col[i](
                tf.expand_dims(inputs[:, :, i, :], axis=-2)))
        return tf.concat(convolutions, axis=-2)


# k-max pooling
class KMaxPooling(layers.Layer):
    """
    KMaxPooling Layer: a pooling operation that is a generalisation
    of the max pooling over the time dimension used in the Max-TDNN
    sentence model and different from the local max pooling operations
    applied in a convolutional network for object recognition
    (LeCun et al., 1998).

    Parameters
    ----------
        ktop: int
            is the fixed pooling parameter for the
            topmost convolutional layer (Sect. 3.2 - https://arxiv.org/pdf/1404.2188.pdf)
            For instance, in a network with three convolutional layers and ktop = 3

        L: int
            L is the total number of convolutional layers in the network

        l: int
            l is the number of the current convolutional
            layer to which the pooling is applied
    """
    def __init__(self, ktop=4, L=None, l=None):
        super(KMaxPooling, self).__init__()
        self.ktop = ktop
        self.L = L
        self.l = l

    def build(self, input_shape):
        s = input_shape[1]
        if self.L is None or self.l is None or s is None:
            self.k = self.ktop
        else:
            self.k = max(self.ktop, math.ceil((self.L-self.l)/self.L*s))

    def call(self, inputs):
        inputs_trans = tf.transpose(inputs, [0, 3, 2, 1])
        inputs_trans_kmax = tf.math.top_k(inputs_trans, self.k).values
        inputs_kmax = tf.transpose(inputs_trans_kmax, [0, 3, 2, 1])
        return inputs_kmax


class Folding(layers.Layer):
    """
    Folding Layer: In the formulation of the network so far, feature
    detectors applied to an individual row of the sentence matrix s can have many orders and create
    complex dependencies across the same rows in
    multiple feature maps. Feature detectors in different rows, however, are independent of each other
    until the top fully connected layer. Full dependence between different rows could be achieved
    by making M in Eq. 5 a full matrix instead of
    a sparse matrix of diagonals. Here we explore a
    simpler method called folding that does not introduce any additional parameters.
    After a convolutional layer and before (dynamic) k-max pooling, one just
    sums every two rows in a feature map component-wise.
    For a map of d rows, folding returns a map of d/2 rows, thus halving the size of
    the representation. With a folding layer, a feature
    detector of the i-th order depends now on two rows
    of feature values in the lower maps of or

     https://arxiv.org/pdf/1404.2188.pdf

    """
    def __init__(self):
        super(Folding, self).__init__()

    def call(self, inputs):
        folded_inputs = tf.math.add_n(
            [inputs[:, :, 0::2, :], inputs[:, :, 1::2, :]]
        ) / 2
        return folded_inputs


"""
Buid CNN Model
"""


class DCNN_Advanced(tf.keras.Model):
    """
    A Convolutional Neural Network for Modelling Sentences

    The ability to accurately represent sentences is central
    to language understanding. We describe a convolutional
    architecture dubbed the Dynamic Convolutional
    Neural Network (DCNN) that we adopt
    for the semantic modelling of sentences.
    The network uses Dynamic k-Max Pooling, a global pooling
    operation over linear sequences. The network handles input
    sentences of varying length and induces
    a feature graph over the sentence that is
    capable of explicitly capturing short and
    long-range relations. The network does
    not rely on a parse tree and is easily applicable to any language.
    We test the DCNN in four experiments: small scale
    binary and multi-class sentiment prediction, six-way question classification and
    Twitter sentiment prediction by distant supervision. The network achieves excellent
    performance in the first three tasks and a
    greater than 25% error reduction in the last
    task with respect to the strongest baseline

    Parameters
    ----------
        vocab_size:
            Vocabulary size of the algorithm input text.

        emb_dim : int
            Embedding size.

        nb_filters_1 : int
            Filter size for each layer Conv1D.

        conv_width_1 : int
            Convolution 2D size

        ktop_max: int
            is the fixed pooling parameter for the
            topmost convolutional layer (Sect. 3.2 - https://arxiv.org/pdf/1404.2188.pdf)
            For instance, in a network with three convolutional layers and ktop = 3

        nb_filters_2 : int
            Filter size for each layer Conv1D.

        conv_width_2 : int
            Convolution 2D size

        nb_of_layers : int
            Units for dense layer.

        dropout_rate : float
            Dropout parameter.

        nb_classes : int
            Numbers of final categories.

    Attributes
    -----------
        embedding : tf.keras.layers.Embedding
            Embedding layer for input vocabulary.

        widening_1 : Widening custom Layer

        widening_2 : Widening custom Layer

        conv_1 : MyConv2D custom Layer

        conv_2 : MyConv2D custom Layer

        fold: Folding custom Layer

        pool_1 : tf.keras.layers.GlobalMaxPool1D
            Max pooling operation for 1D temporal data.

        pool_2 : tf.keras.layers.GlobalMaxPool1D
            Max pooling operation for 1D temporal data.

        dense : tf.keras.layers.Dense
            Regular densely-connected NN layer, concatenate 1D Convolutions.

        flatten : tf.keras.layers.Flatten

        dropout : tf.keras.layers.Dropout
            Applies Dropout to dense_1.

    Examples:
    ---------
        ```python

        VOCAB_SIZE = tokenizer.vocab_size #66125

        EMB_DIM = 60
        NB_FILTERS_1 = 6
        CONV_WIDTH_1 = 7
        KTOP_MAX = 4
        NB_FILTERS_2 = 14
        CONV_WIDTH_2 = 5
        FOLD_PATCH = 2
        NB_OF_LAYERS = 2
        DROPOUT_RATE = 0.1

        NB_CLASSES = 2

        EPOCHS=5

        Dcnn = DCNN(vocab_size=VOCAB_SIZE,
                    emb_dim=EMB_DIM,
                    nb_filters_1=NB_FILTERS_1,
                    conv_width_1=CONV_WIDTH_1,
                    ktop_max=KTOP_MAX,
                    nb_filters_2=NB_FILTERS_2,
                    conv_width_2=CONV_WIDTH_2,
                    fold_patch=FOLD_PATCH,
                    nb_of_layers=NB_OF_LAYERS,
                    dropout_rate=DROPOUT_RATE,
                    nb_classes=NB_CLASSES)

        if NB_CLASSES == 2:
            Dcnn.compile(loss="binary_crossentropy",
                        optimizer="adagrad",
                        metrics=["accuracy"])
        else:
            Dcnn.compile(loss="sparse_categorical_crossentropy",
                        optimizer="adagrad",
                        metrics=["sparse_categorical_accuracy"])

        Dcnn.fit(train_dataset,
                epochs=EPOCHS
                )

        results = Dcnn.evaluate(test_dataset)
        print(results)
        >>> [0.5345908999443054, 0.7479830384254456]
        ```
    """
    def __init__(self,
                 vocab_size,
                 emb_dim=48,
                 nb_filters_1=6,
                 conv_width_1=7,
                 ktop_max=4,
                 nb_filters_2=14,
                 conv_width_2=5,
                 fold_patch=2,
                 nb_of_layers=2,
                 padding_value=0,
                 dropout_rate=0.1,
                 nb_classes=2,
                 name="dcnn",
                 **kwargs):
        super(DCNN_Advanced, self).__init__(name=name, **kwargs)

        self.embedding = layers.Embedding(vocab_size,
                                          emb_dim)
        self.widening_1 = Widening(conv_width_1-1,
                                   axis=1)
        self.conv_1 = MyConv2D(nb_filters=nb_filters_1,
                               conv_width=conv_width_1,
                               padding="valid",
                               emb_dim=emb_dim)
        self.pool_1 = KMaxPooling(ktop_max,
                                  nb_of_layers,
                                  1)

        self.widening_2 = Widening(conv_width_2-1,
                                   axis=1)
        self.conv_2 = MyConv2D(nb_filters=nb_filters_2,
                               conv_width=conv_width_2,
                               padding="valid",
                               emb_dim=emb_dim)
        #self.fold = Folding(patch_size=fold_patch)
        self.fold = Folding()
        self.pool_2 = KMaxPooling(ktop_max)

        self.dropout = layers.Dropout(rate=dropout_rate)
        self.flatten = layers.Flatten()
        if nb_classes == 2:
            self.dense = layers.Dense(1, activation="sigmoid")
        else:
            self.dense = layers.Dense(nb_classes,
                                      activation="softmax")

    def call(self, inputs, training):
        x = self.embedding(inputs)
        x = tf.expand_dims(x, axis=-1)

        x = self.widening_1(x)
        x = self.conv_1(x)
        x = self.pool_1(x)

        x = self.widening_2(x)
        x = self.conv_2(x)
        x = self.fold(x)
        x = self.pool_2(x)

        x = self.dropout(x, training)
        x = self.flatten(x)
        x = self.dense(x)
        return x



"""
Jaime Sendra Berenguer-2020.
MLearner Machine Learning Library Extensions
Author:Jaime Sendra Berenguer<www.linkedin.com/in/jaisenbe>
License: MIT
"""

import numpy as np
import tensorflow as tf


def keras_checkpoint(model, optimizer, checkpoint_path="", max_to_keep=5):
    """
    Manantiene una save_counter para numerar los puestos de control.

    Parameters
    ----------
        model: tf.keras.Model.
            Input model

        checkpoint_path: str
            Input model

        max_to_keep: int
            Maximo numero de checkpoints guardados

    Returns:
    --------
        ckpt_manager: tf.train.Checkpoint.
            Clase Checkpoint
    """
    ckpt = tf.train.Checkpoint(Model=model, optimizer=optimizer)
    ckpt_manager = tf.train.CheckpointManager(ckpt, checkpoint_path, max_to_keep=max_to_keep)

    if ckpt_manager.latest_checkpoint:
        ckpt.restore(ckpt_manager.latest_checkpoint)
        print("The last checkpoint has been restored")

    return ckpt_manager


class MyCustomCallback(tf.keras.callbacks.Callback):
    """
    Keras Callback
    """
    def __init__(self, checkpoint_path, ckpt_manager):
        super(MyCustomCallback, self).__init__()

        self.checkpoint_path = checkpoint_path
        self.ckpt_manager = ckpt_manager

    def on_epoch_end(self, epoch, logs=None):
        self.ckpt_manager.save()
        print("Checkpoint guardado en {}.".format(self.checkpoint_path))


class EarlyStoppingAtMinLoss(tf.keras.callbacks.Callback):
    """Detener el entrenamiento cuando la perdida (loss) esta en su minimo,
    i.e. la perdida (loss) deja de disminuir.

    Parameters
    -----------
        patience: int
            Numero de epochs a esperar despues de que el min ha sido alcanzaado.
            Despues de este numero de no mejoras, el entrenamiento para.
    """
    def __init__(self, patience=0):
        super(EarlyStoppingAtMinLoss, self).__init__()

        self.patience = patience

        # best_weights para almacenar los pesos en los cuales ocurre la perdida minima.
        self.best_weights = None

    def on_train_begin(self, logs=None):
        # El numero de epoch que ha esperado cuando la perdida ya no es minima.
        self.wait = 0
        # El epoch en el que en entrenamiento se detiene.
        self.stopped_epoch = 0
        # Initialize el best como infinito.
        self.best = np.Inf

    def on_epoch_end(self, epoch, logs=None):
        current = logs.get('loss')
        if np.less(current, self.best):
            self.best = current
            self.wait = 0
            # Guardar los mejores pesos si el resultado actual es mejor (menos).
            self.best_weights = self.model.get_weights()
        else:
            self.wait += 1
            if self.wait >= self.patience:
                self.stopped_epoch = epoch
                self.model.stop_training = True
                print('Restaurando los pesos del modelo del final de la mejor epoch.')
                self.model.set_weights(self.best_weights)

    def on_train_end(self, logs=None):
        if self.stopped_epoch > 0:
            print('Epoch %05d: Detencion anticipada' % (self.stopped_epoch + 1))


class LearningRateScheduler(tf.keras.callbacks.Callback):
    """
    Planificador de Learning rate que define el learning
    rate deacuerdo a lo programado.

    Parameters
    -----------
        schedule:
            una funcion que toma el indice del epoch
            (entero, indexado desde 0) y el learning rate actual
            como entradas y regresa un nuevo learning rate como salida (float).

    Examples:
    ---------
        ```python
        LR_SCHEDULE = [
            # (epoch a comenzar, learning rate) tupla
            (3, 0.05), (6, 0.01), (9, 0.005), (12, 0.001)
        ]

        def lr_schedule(epoch, lr):
        #Funcion de ayuda para recuperar el learning rate programado basado en la epoch.
        if epoch < LR_SCHEDULE[0][0] or epoch > LR_SCHEDULE[-1][0]:
            return lr
        for i in range(len(LR_SCHEDULE)):
            if epoch == LR_SCHEDULE[i][0]:
            return LR_SCHEDULE[i][1]
        return lr

        model = get_model()
        _ = model.fit(x_train, y_train,
                batch_size=64,
                steps_per_epoch=5,
                epochs=15,
                verbose=0,
                callbacks=[LossAndErrorPrintingCallback(), LearningRateScheduler(lr_schedule)])
        ```
    """
    def __init__(self, schedule):
        super(LearningRateScheduler, self).__init__()
        self.schedule = schedule

    def on_epoch_begin(self, epoch, logs=None):
        if not hasattr(self.model.optimizer, 'lr'):
            raise ValueError('Optimizer must have a "lr" attribute.')
        # Obtener el learning rate actua del optimizer del modelo.
        lr = float(tf.keras.backend.get_value(self.model.optimizer.lr))
        # Llamar la funcion schedule para obtener el learning rate programado.
        scheduled_lr = self.schedule(epoch, lr)
        # Definir el valor en el optimized antes de que la epoch comience
        tf.keras.backend.set_value(self.model.optimizer.lr, scheduled_lr)
        print('\nEpoch %05d: Learning rate is %6.4f.' % (epoch, scheduled_lr))




"""Jaime Sendra Berenguer-2020.
MLearner Machine Learning Library Extensions
Author:Jaime Sendra Berenguer<www.linkedin.com/in/jaisenbe>
License: MIT
"""

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.decomposition import PCA
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from sklearn.preprocessing import StandardScaler


class PCA_selector(BaseEstimator, TransformerMixin):
    def __init__(self, columns=None, n_components=2, random_state=99):
        """Init log PCA_selector."""
        if columns is not None:
            if isinstance(columns, list) or isinstance(columns, tuple):
                self.columns = columns
            else:
                raise TypeError("Invalid type {}".format(type(columns)))
        else:
            self.columns = columns

        if n_components is not None:
            if isinstance(n_components, int):
                self.n_components = n_components
            else:
                raise TypeError("Invalid type {}".format(type(n_components)))
        else:
            self.n_components = n_components
        self.random_state = random_state

    def fit(self, X, y=None):
        """Selecting PCA columns from the dataset.

        Parameters
        ----------
        X : {Dataframe}, shape = [n_samples, n_features]
            Dataframe, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        --------
        self
        """
        if self.columns is None:
            self.columns = X.select_dtypes(exclude=["object"]).columns

        if isinstance(X, pd.core.frame.DataFrame):
            try:
                _test = X[self.columns].astype(np.float32)
                del(_test)
            except ValueError:
                raise NameError("Null or categorical variables are not allowed: {}".format(X.dtypes))
        else:
            raise NameError("Invalid type {}".format(type(X)))

        self._fitted = True

        return self

    def transform(self, X):
        """Trransformer applies PCA.

        Parameters
        ----------
        X : {Dataframe}, shape = [n_samples, n_features]
            Dataframe of samples, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        -------
        X_transform : {DAtaframe}, shape = [n_samples, n_features]
            A copy of the input Dataframe with the columns centered.
        """

        if not hasattr(self, "_fitted"):
            raise AttributeError("PCA_selector has not been fitted, yet.")

        if not isinstance(X, pd.core.frame.DataFrame):
            raise NameError("Invalid type {}".format(type(X)))

        self.X_std = StandardScaler().fit_transform(X[self.columns])

        return PCA(n_components=self.n_components, random_state=self.random_state).fit_transform(self.X_std)


class LDA_selector(BaseEstimator, TransformerMixin):
    def __init__(self, columns=None, random_state=99):
        """Init log LDA_selector."""
        if columns is not None:
            if isinstance(columns, list) or isinstance(columns, tuple):
                self.columns = columns
            else:
                raise TypeError("Invalid type {}".format(type(columns)))
        else:
            self.columns = columns
        self.random_state = random_state

    def fit(self, X, y):
        """Selecting LDA columns from the dataset.

        Parameters
        ----------
        X : {Dataframe}, shape = [n_samples, n_features]
            Dataframe, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        --------
        self
        """
        if self.columns is None:
            self.columns = X.select_dtypes(exclude=["object"]).columns

        if isinstance(X, pd.core.frame.DataFrame):
            try:
                _test = X[self.columns].astype(np.float32)
                del(_test)
            except ValueError:
                raise NameError("Null or categorical variables are not allowed: {}".format(X.dtypes))
        else:
            raise NameError("Invalid type {}".format(type(X)))

        if not isinstance(y, pd.core.frame.DataFrame) and not isinstance(y, pd.core.series.Series):
            raise NameError("Invalid type {}".format(type(y)))

        self.LDA = LinearDiscriminantAnalysis().fit(X[self.columns], y)

        return self

    def transform(self, X):
        """Trransformer applies LDA.

        Parameters
        ----------
        X : {Dataframe}, shape = [n_samples, n_features]
            Dataframe of samples, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        -------
        X_transform : {DAtaframe}, shape = [n_samples, n_features]
            A copy of the input Dataframe with the columns centered.
        """

        if not hasattr(self, "LDA"):
            raise AttributeError("LDA_selector has not been fitted, yet.")

        if not isinstance(X, pd.core.frame.DataFrame):
            raise NameError("Invalid type {}".format(type(X)))

        return self.LDA.transform(X[self.columns])


class PCA_add(BaseEstimator, TransformerMixin):
    def __init__(self, columns=None, n_components=2, PCA_name=None, random_state=99):
        """Init log PCA_add."""
        if columns is not None:
            if isinstance(columns, list) or isinstance(columns, tuple):
                self.columns = columns
            else:
                raise TypeError("Invalid type {}".format(type(columns)))
        else:
            self.columns = columns

        if n_components is not None:
            if isinstance(n_components, int):
                self.n_components = n_components
            else:
                raise TypeError("Invalid type {}".format(type(n_components)))
        else:
            self.n_components = n_components

        self.PCA_name = PCA_name
        self.random_state = random_state

    def fit(self, X, y=None):
        """Selecting PCA columns from the dataset.

        Parameters
        ----------
        X : {Dataframe}, shape = [n_samples, n_features]
            Dataframe, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        --------
        self
        """
        if self.columns is None:
            self.columns = X.select_dtypes(exclude=["object"]).columns

        if isinstance(X, pd.core.frame.DataFrame):
            try:
                _test = X[self.columns].astype(np.float32)
                del(_test)
            except ValueError:
                raise NameError("Null or categorical variables are not allowed: {}".format(X.dtypes))
        else:
            raise NameError("Invalid type {}".format(type(X)))

        self._fitted = True

        return self

    def _add_dataframe(self, Y):

        df = pd.DataFrame()
        for i in range(Y.shape[1]):
            nombre = str(self.PCA_name) + "_PCA_" + str(i+1)
            df[nombre] = Y[:, i]
        return df

    def _concat_dataframe(self, X_transf, df):

        for i in df.columns.tolist():
            X_transf[i] = df[i].values
        return X_transf

    def transform(self, X):
        """Trransformer applies PCA.

        Parameters
        ----------
        X : {Dataframe}, shape = [n_samples, n_features]
            Dataframe of samples, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        -------
        X_transform : {DAtaframe}, shape = [n_samples, n_features]
            A copy of the input Dataframe with the columns centered.
        """

        if not hasattr(self, "_fitted"):
            raise AttributeError("PCA_selector has not been fitted, yet.")

        if not isinstance(X, pd.core.frame.DataFrame):
            raise NameError("Invalid type {}".format(type(X)))

        X_transform = X.copy()
        self.X_std = StandardScaler().fit_transform(X_transform[self.columns])
        X_PCA = PCA(n_components=self.n_components, random_state=self.random_state).fit_transform(self.X_std)
        df_PCA = self._add_dataframe(X_PCA)

        return self._concat_dataframe(X_transform, df_PCA)


class LDA_add(BaseEstimator, TransformerMixin):
    def __init__(self, columns=None, LDA_name=None, random_state=99):
        """Init log LDA_selector."""
        if columns is not None:
            if isinstance(columns, list) or isinstance(columns, tuple):
                self.columns = columns
            else:
                raise TypeError("Invalid type {}".format(type(columns)))
        else:
            self.columns = columns

        self.LDA_name = LDA_name
        self.random_state = random_state

    def fit(self, X, y=None):
        """Selecting LDA columns from the dataset.

        Parameters
        ----------
        X : {Dataframe}, shape = [n_samples, n_features]
            Dataframe, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        --------
        self
        """
        if self.columns is None:
            self.columns = X.select_dtypes(exclude=["object"]).columns

        if isinstance(X, pd.core.frame.DataFrame):
            try:
                _test = X[self.columns].astype(np.float32)
                del(_test)
            except ValueError:
                raise NameError("Null or categorical variables are not allowed: {}".format(X.dtypes))
        else:
            raise NameError("Invalid type {}".format(type(X)))

        if not isinstance(y, pd.core.frame.DataFrame) and not isinstance(y, pd.core.series.Series):
            raise NameError("Invalid type {}".format(type(y)))

        self.LDA = LinearDiscriminantAnalysis().fit(X[self.columns], y)

        return self

    def _add_dataframe_LDA(self, Y):

        df_LDA = pd.DataFrame()
        for i in range(Y.shape[1]):
            nombre = str(self.LDA_name) + "_LDA_" + str(i+1)
            df_LDA[nombre] = Y[:, i]
        return df_LDA

    def _concat_dataframe(self, X_transf, df_LDA):

        for i in df_LDA.columns.tolist():
            X_transf[i] = df_LDA[i].values
        return X_transf

    def transform(self, X):
        """Trransformer applies LDA.

        Parameters
        ----------
        X : {Dataframe}, shape = [n_samples, n_features]
            Dataframe of samples, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        -------
        X_transform : {DAtaframe}, shape = [n_samples, n_features]
            A copy of the input Dataframe with the columns centered.
        """

        if not hasattr(self, "X_LDA"):
            raise AttributeError("LDA_selector has not been fitted, yet.")

        if not isinstance(X, pd.core.frame.DataFrame):
            raise NameError("Invalid type {}".format(type(X)))

        X_transform = X.copy()
        X_LDA = self.LDA.transform(X_transform)
        df_LDA = self._add_dataframe_LDA(X_LDA)

        return self._concat_dataframe(X_transform, df_LDA)



"""
Jaime Sendra Berenguer-2020.
MLearner Machine Learning Library Extensions
Author:Jaime Sendra Berenguer<www.linkedin.com/in/jaisenbe>
License: MIT
"""

import tensorflow as tf
from tensorflow.keras import layers

"""
Based on https://www.udemy.com/course/procesamiento-del-lenguaje-natural/learn/lecture/21502260#overview

Author: Juan Gabriel Gomila
Course: "Procesamiento del Lenguaje Natural Moderno en Python.

Original Text: https://arxiv.org/pdf/1404.2188.pdf

"""


class DCNN(tf.keras.Model):
    """
    The DCNN class corresponds to the Neural Convolution Network algorithm
    for Natural Language Processing.

    Parameters
    ----------
        vocab_size:
            Vocabulary size of the algorithm input text.

        emb_dim : int
            Embedding size.

        nb_filters : int
            Filter size for each layer Conv1D.

        FFN_units : int
            Units for dense layer.

        nb_classes : int
            Numbers of final categories.

        dropout_rate : float
            Dropout parameter.

        training : bool
            Trainning process activated.

        name : str
            Custom Model Name.

        weights_path: str
            Path load weight model.

    Attributes
    -----------
        embedding : tf.keras.layers.Embedding
            Embedding layer for input vocabulary.

        bigram : tf.keras.layers.Conv1D
            1D convolution layer, for two letters in a row.

        trigram : tf.keras.layers.Conv1D
            1D convolution layer, for three letters in a row.

        fourgram : tf.keras.layers.Conv1D
            1D convolution layer, for four letters in a row.

        pool : tf.keras.layers.GlobalMaxPool1D
            Max pooling operation for 1D temporal data.

        dense_1 : tf.keras.layers.Dense
            Regular densely-connected NN layer, concatenate 1D Convolutions.

        last_dense : tf.keras.layers.Dense
            Regular densely-connected NN layer, final decision.

        dropout : tf.keras.layers.Dropout
            Applies Dropout to dense_1.

    Examples:
    ---------
        ```python
        VOCAB_SIZE = tokenizer.vocab_size # 65540
        EMB_DIM = 200
        NB_FILTERS = 100
        FFN_UNITS = 256
        NB_CLASSES = 2#len(set(train_labels))
        DROPOUT_RATE = 0.2
        BATCH_SIZE = 32
        NB_EPOCHS = 5

        Dcnn = DCNN(vocab_size=VOCAB_SIZE,
            emb_dim=EMB_DIM,
            nb_filters=NB_FILTERS,
            FFN_units=FFN_UNITS,
            nb_classes=NB_CLASSES,
            dropout_rate=DROPOUT_RATE)

        if NB_CLASSES == 2:
            Dcnn.compile(loss="binary_crossentropy",
                        optimizer="adam",
                        metrics=["accuracy"])
        else:
            Dcnn.compile(loss="sparse_categorical_crossentropy",
                        optimizer="adam",
                        metrics=["sparse_categorical_accuracy"])

        # Entrenamiento
        Dcnn.fit(train_inputs,
                train_labels,
                batch_size=BATCH_SIZE,
                epochs=NB_EPOCHS)

        # Evaluation
        results = Dcnn.evaluate(test_inputs, test_labels, batch_size=BATCH_SIZE)
        print(results)
        ```
    """
    def __init__(self,
                 vocab_size,
                 emb_dim=128,
                 nb_filters=50,
                 FFN_units=512,
                 nb_classes=2,
                 dropout_rate=0.1,
                 training=False,
                 weights_path=None,
                 name="dcnn"):
        super(DCNN, self).__init__(name=name)

        self.weights_path = weights_path

        self.embedding = layers.Embedding(vocab_size,
                                          emb_dim)
        self.bigram = layers.Conv1D(filters=nb_filters,
                                    kernel_size=2,
                                    padding="valid",
                                    activation="relu")
        self.trigram = layers.Conv1D(filters=nb_filters,
                                     kernel_size=3,
                                     padding="valid",
                                     activation="relu")
        self.fourgram = layers.Conv1D(filters=nb_filters,
                                      kernel_size=4,
                                      padding="valid",
                                      activation="relu")
        # No tenemos variable de entrenamiento
        # asi que podemos usar la misma capa
        # para cada paso de pooling
        self.dense_1 = layers.Dense(units=FFN_units, activation="relu")
        self.pool = layers.GlobalMaxPool1D()
        self.dropout = layers.Dropout(rate=dropout_rate)
        if nb_classes == 2:
            self.last_dense = layers.Dense(units=1,
                                           activation="sigmoid")
        else:
            self.last_dense = layers.Dense(units=nb_classes,
                                           activation="softmax")

    def call(self, inputs, training):
        """
        Calling the build function of the model.

        Parameters
        ----------
            inputs: Tensor.
                Input Tensor.

            Training : bool
                Trainning process activated.

        Returns:
            output: Tensor.
                Output Tensor.
        """
        x = self.embedding(inputs)
        x_1 = self.bigram(x)
        x_1 = self.pool(x_1)
        x_2 = self.trigram(x)
        x_2 = self.pool(x_2)
        x_3 = self.fourgram(x)
        x_3 = self.pool(x_3)

        merged = tf.concat([x_1, x_2, x_3], axis=-1)  # (batch_size, 3 * nb_filters)
        merged = self.dense_1(merged)
        merged = self.dropout(merged, training)
        output = self.last_dense(merged)

        # Load weights.
        if self.weights_path is not None:
            output.load_weights(self.weights_path)

        return output



# from distutils.core import setup
from os.path import realpath, dirname, join
from setuptools import setup, find_packages
import mlearner

VERSION = mlearner.__version__
PROJECT_ROOT = dirname(realpath(__file__))

REQUIREMENTS_FILE = join(PROJECT_ROOT, 'requirements.txt')

with open(REQUIREMENTS_FILE) as f:
    install_reqs = f.read().splitlines()

install_reqs.append('setuptools')


def readme():
    with open("README.md", "r") as fh:
        return fh.read()


setup(
  name='mlearner',         # How you named your package folder (MyLib)
  packages=find_packages(),   # Chose the same as "name"
  version=VERSION,      # Start with a small number and increase it with every change you make
  license='MIT',        # Chose a license from here: https://help.github.com/articles/licensing-a-repository
  description='Machine Learning Library Extensions',   # Give a short description about your library
  author='Jaime Sendra',                   # Type in your name
  author_email='jaisenberafel@gmail.com',      # Type in your E-Mail
  url='https://github.com/jaisenbe58r/MLearner',   # Provide either the link to your github or to your website
  download_url='https://github.com/jaisenbe58r/MLearner/archive/v0.0.2.tar.gz',    # I explain this later on
  keywords=['SOME', 'MEANINGFULL', 'KEYWORDS'],   # Keywords that define your package best
  package_data={'': [
                    'README.md',
                    'requirements.txt']
                },
  include_package_data=True,
  install_requires=install_reqs,
  platforms='any',

  classifiers=[
    'Development Status :: 3 - Alpha',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
    'Intended Audience :: Developers',      # Define that your audience are developers
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',   # Again, pick a license
    'Programming Language :: Python :: 3',      # Specify which pyhton versions that you want to support
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Operating System :: Microsoft :: Windows',
    'Operating System :: POSIX',
    'Operating System :: Unix',
    'Operating System :: MacOS',
    'Programming Language :: Python :: 3.7',
    'Topic :: Scientific/Engineering',
    'Topic :: Scientific/Engineering :: Artificial Intelligence',
    'Topic :: Scientific/Engineering :: Information Analysis',
    'Topic :: Scientific/Engineering :: Image Recognition',
  ],
    python_requires='>=3.6',
    long_description=readme(),
    long_description_content_type="text/markdown",
)

# setup(
#   name = 'mlearner',         # How you named your package folder (MyLib)
#   packages = ['mlearner'],   # Chose the same as "name"
#   version = VERSION,      # Start with a small number and increase it with every change you make
#   license='MIT',        # Chose a license from here: https://help.github.com/articles/licensing-a-repository
#   description = 'Machine Learning Library Extensions',   # Give a short description about your library
#   author = 'Jaime Sendra Berenguer',                   # Type in your name
#   author_email = 'jaisenberafel@gmail.com',      # Type in your E-Mail
#   url = 'https://github.com/jaisenbe58r/MLearner',   # Provide either the link to your github or to your website
#   download_url = 'https://github.com/jaisenbe58r/MLearner/archive/v0.0.2.tar.gz',    # I explain this later on
#   keywords = ['SOME', 'MEANINGFULL', 'KEYWORDS'],   # Keywords that define your package best
#   install_requires=[            # I get to this in a second
#           'validators',
#           'beautifulsoup4',
#       ],
#   classifiers=[
#     'Development Status :: 3 - Alpha',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
#     'Intended Audience :: Developers',      # Define that your audience are developers
#     'Topic :: Software Development :: Build Tools',
#     'License :: OSI Approved :: MIT License',   # Again, pick a license
#     'Programming Language :: Python :: 3',      #Specify which pyhton versions that you want to support
#     'Programming Language :: Python :: 3.4',
#     'Programming Language :: Python :: 3.5',
#     'Programming Language :: Python :: 3.6',
#     ],
#     python_requires='>=3.6',
#     ong_description=long_description,
#     long_description_content_type="text/markdown",
# )